(*
    LCS - A study for a simple differencing program
    Open files window
    Copyright (c) 2002-2003,2004 Alessandro Scotti
    www.ascotti.org

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)
unit FormOpenFiles;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls;

type
  TOpenForm = class(TForm)
    LblLeftFile: TLabel;
    BtnBrowseLeft: TButton;
    LblRightFile: TLabel;
    BtnBrowseRight: TButton;
    OpenDlgLeft: TOpenDialog;
    OpenDlgRight: TOpenDialog;
    BtnCancel: TButton;
    BtnOk: TButton;
    CboLeft: TComboBox;
    CboRight: TComboBox;
    ImageOk1: TImage;
    ImageBad1: TImage;
    ImageBad2: TImage;
    ImageOk2: TImage;
    procedure BtnBrowseLeftClick(Sender: TObject);
    procedure BtnBrowseRightClick(Sender: TObject);
    procedure BtnOkClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure CboLeftChange(Sender: TObject);
    procedure CboRightChange(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormShow(Sender: TObject);
  private
    { Private declarations }
    FLeftFile: string;
    FRightFile: string;
    FRightFocus: boolean;
    FAutoOpen: boolean;

    procedure AddToHistory( Box: TComboBox );
    procedure SetLeftFile( AName: string );
    procedure SetRightFile( AName: string );
    procedure OnDropFiles( var M: TMessage ); message WM_DROPFILES;
  public
    { Public declarations }
    property LeftFile: string read FLeftFile write SetLeftFile;
    property RightFile: string read FRightFile write SetRightFile;

    procedure SetFocusToRightFile( F: boolean );
    procedure SetFileFilters( Filters: string );
  end;

var
  OpenForm: TOpenForm;

implementation

uses
  ShellAPI, FileCtrl, ActiveX, ShlObj, Masks,
  Utils;

{$R *.DFM}

const
  MaxHistoryLength = 8;

procedure TOpenForm.AddToHistory( Box: TComboBox );
var
  I: integer;
begin
  if( Box.Text <> '' ) then begin
    I := Box.Items.IndexOf( Box.Text );
    if( I >= 0 ) then
      Box.Items.Move( I, 0 )
    else
      Box.Items.Insert( 0, Box.Text );
  end;
  if( Box.Items.Count > MaxHistoryLength ) then begin
    Box.Items.Delete( Box.Items.Count-1 );
  end;
end;

function GetIndexOfBestFilter( Filter: string; Filename: string ): integer;
var
  Desc, Mask: string;
begin
  Result := 1;
  repeat
    Desc := ExtractToken( Filter, '|' );
    Mask := ExtractToken( Filter, '|' );
    while( Mask <> '' ) do begin
      if( MatchesMask( FileName, ExtractToken(Mask,';') ) ) then
        Exit;
    end;
    Inc( Result );
  until( Filter = '' );
  Result := 1; // Not found
end;

function ChooseFile( OD: TOpenDialog; LF: string ): boolean;
var
  D: string;
begin
  D := ExcludeTrailingBackslash( ExtractFilePath(LF) );
  if( DirectoryExists( D ) ) then
    OD.InitialDir := D
  else
    OD.InitialDir := GetCurrentDir;
  OD.FilterIndex := GetIndexOfBestFilter( OD.Filter, LF );
  Result := OD.Execute;
end;

procedure TOpenForm.BtnBrowseLeftClick(Sender: TObject);
begin
  if( ChooseFile( OpenDlgLeft, CboLeft.Text ) ) then
    SetLeftFile( OpenDlgLeft.FileName );
  CboLeft.SelStart := Length( CboLeft.Text );
end;

procedure TOpenForm.BtnBrowseRightClick(Sender: TObject);
begin
  if( ChooseFile( OpenDlgRight, CboRight.Text ) ) then
    SetRightFile( OpenDlgRight.FileName );
  CboRight.SelStart := Length( CboRight.Text );
end;

procedure TOpenForm.BtnOkClick(Sender: TObject);
begin
  AddToHistory( CboLeft );
  AddToHistory( CboRight );
end;

procedure TOpenForm.FormCreate(Sender: TObject);
begin
  Constraints.MinHeight := Height;
  Constraints.MaxHeight := Height;
  FLeftFile := 'n/a';
  SetLeftFile( '' );
  FRightFile := 'n/a';
  FRightFocus := false;
  FAutoOpen := false;
  SetRightFile( '' );

  DragAcceptFiles( Handle, true );
end;

procedure TOpenForm.SetFileFilters( Filters: string );
begin
  OpenDlgLeft.Filter := Filters;
  OpenDlgRight.Filter := Filters;
end;

procedure TOpenForm.SetLeftFile( AName: string );
var
  F: boolean;
begin
  if( CboLeft.Text <> AName ) then
    CboLeft.Text := AName;

  if( FLeftFile <> AName ) then begin
    FLeftFile := AName;
    F := FileExists(AName);
    ImageBad1.Visible := not F;
    ImageOk1.Visible := F;
    BtnOk.Enabled := ImageOk1.Visible and ImageOk2.Visible;
  end;
end;

procedure TOpenForm.SetRightFile( AName: string );
var
  F: boolean;
begin
  if( CboRight.Text <> AName ) then
    CboRight.Text := AName;

  if( FRightFile <> AName ) then begin
    FRightFile := AName;
    F := FileExists(AName);
    ImageBad2.Visible := not F;
    ImageOk2.Visible := F;
    BtnOk.Enabled := ImageOk1.Visible and ImageOk2.Visible;
  end;
end;

procedure TOpenForm.CboLeftChange(Sender: TObject);
begin
  SetLeftFile( CboLeft.Text );
end;

procedure TOpenForm.CboRightChange(Sender: TObject);
begin
  SetRightFile( CboRight.Text );
end;

// Handles file dragged from another application
procedure TOpenForm.OnDropFiles( var M: TMessage );
var
  Drop: THandle;
  Name: array[ 1..MAX_PATH ] of char;
  N: integer;
  DW: DWORD;
  P: TPoint;
begin
  Drop := M.WParam;
  N := DragQueryFile( Drop, $FFFFFFFF, nil, 0 );
  if( N > 0 ) then begin
    // Get name of first file
    DragQueryFile( Drop, 0, @Name, sizeof(Name) );
    if( N = 1 ) then begin
      // Only one file, select destination according to drop position
      DW := GetMessagePos;
      P.X := LOWORD(DW);
      P.Y := HIWORD(DW);
      P := ScreenToClient( P );
      if( PtInRect( CboRight.BoundsRect, P ) ) then
        SetRightFile( StrPas(@Name) )
      else
        SetLeftFile( StrPas(@Name) );
    end
    else begin
      // Two or more files, use first two names
      SetLeftFile( StrPas(@Name) );
      DragQueryFile( Drop, 1, @Name, sizeof(Name) );
      SetRightFile( StrPas(@Name) );
    end;
  end;
  DragFinish( Drop );
  M.Result := 0;
end;

procedure TOpenForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if( Shift <> [ssCtrl] ) then
    Exit;

  case Key of
    Ord('L'):
      BtnBrowseLeft.Click;
    Ord('R'):
      BtnBrowseRight.Click;
    Ord('A'):
      // Select all
      if( ActiveControl = BtnBrowseLeft )or( ActiveControl = CboLeft ) then begin
        CboLeft.SelStart := 0;
        CboLeft.SelLength := Length(CboLeft.Text);
      end
      else if( ActiveControl = BtnBrowseRight )or( ActiveControl = CboRight ) then begin
        CboRight.SelStart := 0;
        CboRight.SelLength := Length(CboRight.Text);
      end;
    Ord('O'):
      // Open file choose dialog
      if( ActiveControl = BtnBrowseLeft )or( ActiveControl = CboLeft ) then
        BtnBrowseLeft.Click
      else if( ActiveControl = BtnBrowseRight )or( ActiveControl = CboRight ) then
        BtnBrowseRight.Click;
  end;
end;

procedure TOpenForm.SetFocusToRightFile( F: boolean );
begin
  FRightFocus := F;
end;

procedure TOpenForm.FormShow(Sender: TObject);
begin
  if( FRightFocus ) then begin
    CboRight.SetFocus;
    BtnBrowseRightClick( Self );
  end
  else begin
    CboLeft.SetFocus;
    BtnBrowseLeftClick( Self );
  end;
end;

end.
