(*
    LCS - A study for a simple differencing program
    Main window
    Copyright (c) 2002-2003,2004 Alessandro Scotti
    www.ascotti.org

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)
unit Main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, Menus, ToolWin, ComCtrls,
  DiffTypes, DiffEngine, DiffUndoBuffer, ImgList,
  Config, Options, UndoRedo, Syntax,
  FormOptions;

type
  TChangeBlockAction = (
    baCopyToDoc2, baCopySingleToDoc2,
    baCopyToDoc1, baCopySingleToDoc1,
    baDeleteFromDoc1, baDeleteSingleFromDoc1,
    baDeleteFromDoc2, baDeleteSingleFromDoc2 );

type
  TMainForm = class(TForm)
    StatusBar: TStatusBar;
    ToolBar: TToolBar;
    PbxOverview: TPaintBox;
    MainMenu: TMainMenu;
    CmdFile: TMenuItem;
    CmdOpen: TMenuItem;
    N1: TMenuItem;
    CmdExit: TMenuItem;
    CmdSaveLeftFile: TMenuItem;
    CmdSaveLeftFileAs: TMenuItem;
    CmdSaveRightFile: TMenuItem;
    CmdSaveRightFileAs: TMenuItem;
    PopupLeft: TPopupMenu;
    PopCopyToRight: TMenuItem;
    PopDeleteBlockLeft: TMenuItem;
    N2: TMenuItem;
    PopSaveLeftFile: TMenuItem;
    PopSaveLeftFileAs: TMenuItem;
    PbxView1: TPaintBox;
    Splitter: TSplitter;
    PbxView2: TPaintBox;
    VertScrollBar: TScrollBar;
    PnlBottom: TPanel;
    HorzScrollBar: TScrollBar;
    N3: TMenuItem;
    CmdEdit: TMenuItem;
    ImageList: TImageList;
    BtnOpen: TToolButton;
    ToolButton2: TToolButton;
    BtnSaveAll: TToolButton;
    BtnSaveLeft: TToolButton;
    BtnSaveRight: TToolButton;
    PopupRight: TPopupMenu;
    PopCopyLinesToLeft: TMenuItem;
    PopDeleteBlockRight: TMenuItem;
    N4: TMenuItem;
    PopSaveRightFile: TMenuItem;
    PopSaveRightFileAs: TMenuItem;
    CmdHelp: TMenuItem;
    CmdAbout: TMenuItem;
    Bevel1: TBevel;
    ToolButton6: TToolButton;
    BtnUndo: TToolButton;
    BtnRedo: TToolButton;
    ToolButton9: TToolButton;
    BtnPrevDiff: TToolButton;
    CmdSearch: TMenuItem;
    CmdFind: TMenuItem;
    CmdFindAgain: TMenuItem;
    N5: TMenuItem;
    CmdNextDifference: TMenuItem;
    CmdPreviousDifference: TMenuItem;
    CmdUndo: TMenuItem;
    CmdRedo: TMenuItem;
    FindDialog: TFindDialog;
    SaveDialogRight: TSaveDialog;
    SaveDialogLeft: TSaveDialog;
    PopCopySingleLineToRight: TMenuItem;
    PopCopySingleLineToLeft: TMenuItem;
    N6: TMenuItem;
    CmdPreferences: TMenuItem;
    N7: TMenuItem;
    CmdAutosynch: TMenuItem;
    N8: TMenuItem;
    CmdRefresh: TMenuItem;
    ToolButton1: TToolButton;
    BtnRefresh: TToolButton;
    N9: TMenuItem;
    PopPropertiesLeft: TMenuItem;
    N10: TMenuItem;
    PopPropertiesRight: TMenuItem;
    PopDeleteLineLeft: TMenuItem;
    PopDeleteLineRight: TMenuItem;
    BtnShowLineNumbers: TToolButton;
    ToolButton3: TToolButton;
    BtnNextDiff: TToolButton;
    CmdOpenHelp: TMenuItem;
    N11: TMenuItem;
    BtnSyntaxHilight: TToolButton;
    BtnShowSeparators: TToolButton;
    CmdSaveAll: TMenuItem;
    ToolButton4: TToolButton;
    BtnMerge: TToolButton;
    BtnFind: TToolButton;
    BtnFindAgain: TToolButton;
    ToolButton8: TToolButton;
    N12: TMenuItem;
    CmdShortcuts: TMenuItem;
    N13: TMenuItem;
    PopEditLeft: TMenuItem;
    N14: TMenuItem;
    PopEditRight: TMenuItem;
    PopOpenLeftFile: TMenuItem;
    PopOpenRightFile: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure ScrollBarChange(Sender: TObject);
    procedure CmdOpenClick(Sender: TObject);
    procedure CmdExitClick(Sender: TObject);
    procedure PbxView1MouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PopCopyToRightClick(Sender: TObject);
    procedure PopDeleteBlockLeftClick(Sender: TObject);
    procedure PopDeleteBlockRightClick(Sender: TObject);
    procedure PopCopyLinesToLeftClick(Sender: TObject);
    procedure PbxView2MouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure CmdAboutClick(Sender: TObject);
    procedure PbxOverviewMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseWheel(Sender: TObject; Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
    procedure PbxOverviewMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PbxOverviewMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure SplitterMoved(Sender: TObject);
    procedure ActionUndoExecute(Sender: TObject);
    procedure ActionRedoExecute(Sender: TObject);
    procedure PbxView1StartDrag(Sender: TObject;
      var DragObject: TDragObject);
    procedure PbxView1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PbxView2DragOver(Sender, Source: TObject; X, Y: Integer;
      State: TDragState; var Accept: Boolean);
    procedure PbxView2DragDrop(Sender, Source: TObject; X, Y: Integer);
    procedure PbxView2MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PbxView2StartDrag(Sender: TObject;
      var DragObject: TDragObject);
    procedure PbxView1DragDrop(Sender, Source: TObject; X, Y: Integer);
    procedure PbxView1DragOver(Sender, Source: TObject; X, Y: Integer;
      State: TDragState; var Accept: Boolean);
    procedure CmdSaveLeftFileClick(Sender: TObject);
    procedure CmdSaveRightFileClick(Sender: TObject);
    procedure BtnSaveAllClick(Sender: TObject);
    procedure CmdPreviousDifferenceClick(Sender: TObject);
    procedure CmdNextDifferenceClick(Sender: TObject);
    procedure CmdFindAgainClick(Sender: TObject);
    procedure FindDialogFind(Sender: TObject);
    procedure CmdFindClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure CmdSaveLeftFileAsClick(Sender: TObject);
    procedure CmdSaveRightFileAsClick(Sender: TObject);
    procedure PbxView1DblClick(Sender: TObject);
    procedure PbxView2DblClick(Sender: TObject);
    procedure PopCopySingleLineToRightClick(Sender: TObject);
    procedure PopCopySingleLineToLeftClick(Sender: TObject);
    procedure CmdPreferencesClick(Sender: TObject);
    procedure CmdRecentFilesClick(Sender: TObject);
    procedure CmdRefreshClick(Sender: TObject);
    procedure CmdAutosynchClick(Sender: TObject);
    procedure PopPropertiesLeftClick(Sender: TObject);
    procedure PopDeleteLineRightClick(Sender: TObject);
    procedure PopDeleteLineLeftClick(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure BtnShowLineNumbersClick(Sender: TObject);
    procedure PopPropertiesRightClick(Sender: TObject);
    procedure PbxView2MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure CmdOpenHelpClick(Sender: TObject);
    procedure BtnShowSeparatorsClick(Sender: TObject);
    procedure BtnSyntaxHilightClick(Sender: TObject);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure CmdShortcutsClick(Sender: TObject);
    procedure PopEditLeftClick(Sender: TObject);
    procedure PopEditRightClick(Sender: TObject);
  private
    { Private declarations }
    FDiffDocument: TDiffDocument; // Main document
    FSyntaxDoc: array [0..1] of TList;
    FSyntaxTokenizer: array[0..1] of TSyntaxTokenizer;
    FSyntaxLanguages: TList;
    FFileName1: string; // Name of first file
    FFileName2: string; // Name of second file
    FCurrentLine: integer; // Index of "current" line (for searches etc.)
    FOverviewBackground: TBitmap;
    FSizeView1: integer; // Size of left panel as percentage of client area
    FDefaultCaption: string; // Default form caption (grabbed at program startup)
    FUndoRedoBuffer: TUndoRedoBuffer;
    FFlags: integer; // Modification flags (if some doc has changed etc.)
    FSearchString: string; // Last search string
    FOptions: TAppOptions; // Application options (saved to settings file)
    FOptionsManager: TConfigIniFile;
    FLineNumberLength: integer;
    FHilightSyntax: boolean;
    FOverviewDiffsAreSparse: boolean;

    function AllowCloseCurrentDocument: boolean;
    function EditDoc( M: TDiffLineMode ): boolean;
    function GetActualDocument( Mode: TDiffLineMode ): TStringList;
    function GetBlockEnd( L: integer ): integer;
    function GetBlockStart( L: integer ): integer;
    function GetLineFromOverview( Y: integer ): integer;
    function GetLineHeight: integer;
    function GetDefaultOverviewRect: TRect;
    function GetOverviewThumbRect: TRect;
    function GetVisibleLines: integer;
    function HasDiffDocument: boolean;
    function IndexOfPrevDifference: integer;
    function IndexOfNextDifference: integer;
    function LoadFiles( F1, F2: string ): boolean;
    function RewriteFile( Filename: string; var F: TextFile ): integer;
    function SaveFile( Filename: string; Mode: TDiffLineMode ): boolean;
    function ValidateDragOver( Source: TObject; Y: integer ): boolean;
    procedure ChangeLineBlock( Line: integer; ModeFrom, ModeTo: TDiffLineMode; Block: boolean );
    procedure ClearSyntaxDocuments;
    procedure DiffLineListChanged;
    procedure DisplayHint(Sender: TObject);
    procedure DrawOverviewBackground( C: TCanvas; var R: TRect );
    procedure ExecuteDragDrop( Source: TObject; SourceView: integer; ModeFrom, ModeTo: TDiffLineMode );
    procedure ExecuteUndoEditDoc( Item: TEditDiffDocUndoItem; IsRedo: boolean );
    procedure ExecuteUndoRedo( Item: TDiffUndoItem; NewMode: TDiffLineMode );
    procedure FindNext;
    procedure HandleDoubleClick( Sender: TPaintBox; ModeFrom, ModeTo: TDiffLineMode );
    procedure LoadSyntaxHighlightOptions;
    procedure ModificationFlagsChanged;
    procedure OnDropFiles( var M: TMessage ); message WM_DROPFILES;
    procedure OnRedoEvent( Sender: TObject; Item: TUndoRedoItem );
    procedure OnUndoEvent( Sender: TObject; Item: TUndoRedoItem );
    procedure SetViewPosition( L: integer ); overload;
    procedure SetViewPosition( L: integer; DontNotify: boolean ); overload;
    procedure ShowFileProperties( SelectRightFile: boolean );
    procedure UpdateDifferenceNavigators;
    procedure UpdateFormCaption;
    procedure UpdateHistoryRecent;
    procedure UpdateRecentFilesMenu;
    procedure UpdateSyntaxDocuments;

  public
    { Public declarations }
    procedure ChangeBlock( L: integer; Action: TChangeBlockAction );
  end;

var
  MainForm: TMainForm;

implementation

uses
  IniFiles, Math, ShellApi,
  Utils,
  FormOpenFiles, FormAbout, FormProperties, FormAutosynch, DiffAutoPopup,
  FormEditor;

{$R *.DFM}

const
  FlagModifiedFile1 = $01;
  FlagModifiedFile2 = $02;

  MaxRecentFilesCount = 8;

  MinLineNumberLength = 3;

  S_SyntaxHighlight = 'SyntaxHighlight';

procedure TMainForm.LoadSyntaxHighlightOptions;
var
  L: TConfigString;
  S: string;
  N: string;
  F: TMemIniFile;
  P: integer;
  O: TLanguageOptions;
begin
  L := TConfigString.Create( 'KnownLanguages', '' );
  F := TMemIniFile.Create( FOptionsManager.FileName );
  try
    FOptionsManager.LoadItem( F, S_SyntaxHighlight, L );
    S := L.Value + ',';
    while( S <> '' ) do begin
      P := Pos( ',', S );
      if( P = 0 ) then
        break;
      N := Trim( Copy( S, 1, P-1 ) );
      Delete( S, 1, P );
      if( N <> '' ) then begin
        O := TLanguageOptions.Create( N );
        FOptionsManager.LoadItem( F, S_SyntaxHighlight, O.ConfigGroup );
        if( O.IsValid ) then begin
          // Add sentinels for easier handling
          O.Extensions.Value := LowerCase( ' ' + O.Extensions.Value + ' ' );
          FSyntaxLanguages.Add( O );
        end
        else begin
          O.Free;
        end;
      end;
    end;
  finally
    F.Free;
    L.Free;
  end;
end;

procedure TMainForm.FormCreate(Sender: TObject);
var
  I: integer;
  S, F1, F2: string;
begin
  FCurrentLine := -1;
  FDefaultCaption := Caption;
  FSizeView1 := 50;
  FOverviewBackground := TBitmap.Create;
  FUndoRedoBuffer := TUndoRedoBuffer.Create( UnlimitedUndoRedo );
  FUndoRedoBuffer.MultipleEventForItemList := true;
  FUndoRedoBuffer.OnUndoItem := OnUndoEvent;
  FUndoRedoBuffer.OnRedoItem := OnRedoEvent;
  FSearchString := '';
  FFileName1 := '';
  FFileName2 := '';
  FLineNumberLength := MinLineNumberLength;
  FSyntaxDoc[0] := TList.Create;
  FSyntaxDoc[1] := TList.Create;
  FSyntaxTokenizer[0] := nil;
  FSyntaxTokenizer[1] := nil;
  FSyntaxLanguages := TList.Create;
  FHilightSyntax := true;
  FOverviewDiffsAreSparse := true;

  // Main document
  FDiffDocument := TDiffDocument.Create;
  FDiffDocument.RawDoc1 := TStringList.Create;
  FDiffDocument.RawDoc2 := TStringList.Create;
  FDiffDocument.DiffEngine := TDiffEngine.Create;

  // Persistent options
  FOptions.IgnoreBlanks := TConfigBoolean.Create( 'IgnoreBlanks', true );
  FOptions.IgnoreCase := TConfigBoolean.Create( 'IgnoreCase', false );
  FOptions.TabSize := TConfigInteger.Create( 'TabSize', DefaultTabSize );
  FOptions.BackupMode := TConfigInteger.Create( 'BackupMode', 0 );
  FOptions.DblClickOnRightFileDeletes := TConfigBoolean.Create( 'RightDblClickDeletes', false );
  FOptions.ShowLineNumbers := TConfigBoolean.Create( 'ShowLineNumbers', false );
  FOptions.PopupMargin := TConfigInteger.Create( 'PopupMargin', DefaultPopupMargin );
  FOptions.ColorLineNumber := TConfigInteger.Create( 'ColorLineNumber', DefaultColorLineNumber );
  FOptions.ColorOnlyDoc1 := TConfigInteger.Create( 'ColorOnlyDoc1', DefaultColorOnlyDoc1 );
  FOptions.ColorOnlyDoc2 := TConfigInteger.Create( 'ColorOnlyDoc2', DefaultColorOnlyDoc2 );
  FOptions.ColorDefault := TConfigInteger.Create( 'ColorDefault', DefaultColorDefault );
  FOptions.ColorMissingLine := TConfigInteger.Create( 'ColorMissingLine', DefaultColorMissingLine );
  FOptions.ColorChangedLine := TConfigInteger.Create( 'ColorChangedLine', DefaultColorChangedLine );
  FOptions.OpenFileFilters := TConfigString.Create( 'OpenFileFilters', 'C/C++ source file (*.c;*.cpp;*.cxx;*.h;*.hpp;*.hxx)|*.c;*.cpp;*.cxx;*.h;*.hpp;*.hxx|Delphi/Pascal source file (*.pas;*.dpr)|*.pas;*.dpr|HTML file (*.htm;*.html)|*.htm;*.html|Java source file (*.java)|*.java|Text file (*.txt)|*.txt|All files (*.*)|*.*' );
  FOptions.RestoreWindowSettings := TConfigBoolean.Create( 'RestoreWindowSettings', false );
  FOptions.ShowBlockSeparators := TConfigBoolean.Create( 'BlockSeparators', true );
  FOptions.SyntaxHighlight := TConfigBoolean.Create( 'SyntaxHighlight', false );
  FOptions.WindowPosX := TConfigInteger.Create( 'WindowPosX', -1 );
  FOptions.WindowPosY := TConfigInteger.Create( 'WindowPosY', -1 );
  FOptions.WindowWidth := TConfigInteger.Create( 'WindowWidth', -1 );
  FOptions.WindowHeight := TConfigInteger.Create( 'WindowHeight', -1 );
  FOptions.WindowIsMaximized := TConfigBoolean.Create( 'WindowMaximized', false );
  FOptions.HistoryFile1 := TConfigStringList.Create( 'HistoryFile1' );
  FOptions.HistoryFile2 := TConfigStringList.Create( 'HistoryFile2' );
  FOptions.HistoryRecent := TConfigStringList.Create( 'Recent' );

  FOptionsManager := TConfigIniFile.Create( ChangeFileExt( Application.ExeName, '.ini' ) );
  FOptionsManager.Add( FOptions.IgnoreBlanks );
  FOptionsManager.Add( FOptions.IgnoreCase );
  FOptionsManager.Add( FOptions.RestoreWindowSettings );
  FOptionsManager.Add( FOptions.TabSize );
  FOptionsManager.Add( FOptions.BackupMode );
  FOptionsManager.Add( FOptions.DblClickOnRightFileDeletes );
  FOptionsManager.Add( FOptions.ShowLineNumbers );
  FOptionsManager.Add( FOptions.PopupMargin );
  FOptionsManager.Add( FOptions.HistoryFile1 );
  FOptionsManager.Add( FOptions.HistoryFile2 );
  FOptionsManager.Add( FOptions.HistoryRecent );
  FOptionsManager.Add( FOptions.ColorLineNumber );
  FOptionsManager.Add( FOptions.ColorOnlyDoc1 );
  FOptionsManager.Add( FOptions.ColorOnlyDoc2 );
  FOptionsManager.Add( FOptions.ColorDefault );
  FOptionsManager.Add( FOptions.ColorMissingLine );
  FOptionsManager.Add( FOptions.ColorChangedLine );
  FOptionsManager.Add( FOptions.WindowPosX );
  FOptionsManager.Add( FOptions.WindowPosY );
  FOptionsManager.Add( FOptions.WindowWidth );
  FOptionsManager.Add( FOptions.WindowHeight );
  FOptionsManager.Add( FOptions.WindowIsMaximized );
  FOptionsManager.Add( FOptions.SyntaxHighlight );
  FOptionsManager.Add( FOptions.ShowBlockSeparators );

  // Window
  CmdUndo.Hint := BtnUndo.Hint;
  CmdRedo.Hint := BtnRedo.Hint;
  CmdRefresh.Hint := BtnRefresh.Hint;
  CmdAutosynch.Hint := BtnMerge.Hint;
  CmdOpen.Hint := BtnOpen.Hint;
  CmdSaveAll.Hint := BtnSaveAll.Hint;
  CmdSaveLeftFile.Hint := BtnSaveLeft.Hint;
  CmdSaveRightFile.Hint := BtnSaveRight.Hint;
  CmdFind.Hint := BtnFind.Hint;
  CmdFindAgain.Hint := BtnFindAgain.Hint;

  Application.OnHint := DisplayHint;

  DragAcceptFiles( Handle, true );

  // Load options
  FOptionsManager.Load;

  LoadSyntaxHighlightOptions;

  // Update settings that depend on options
  UpdateRecentFilesMenu;

  if( FOptions.RestoreWindowSettings.Value ) then begin
    if( FOptions.WindowPosX.Value >= 0 )and( FOptions.WindowPosY.Value >= 0 )and( FOptions.WindowWidth.Value > 0 )and( FOptions.WindowHeight.Value > 0 ) then begin
      BoundsRect := Rect(FOptions.WindowPosX.Value,
        FOptions.WindowPosY.Value,
        FOptions.WindowPosX.Value + FOptions.WindowWidth.Value,
        FOptions.WindowPosY.Value + FOptions.WindowHeight.Value );
    end;
    if( FOptions.WindowIsMaximized.Value ) then
      WindowState := wsMaximized;
  end;

  // Load files specified on the command line
  F1 := '';
  F2 := '';
  for I:=1 to ParamCount do begin
    S := Trim(ParamStr(I));
    if( S[1] <> '/' ) then begin
      if( F1 = '' ) then
        F1 := ExpandFileName(S)
      else if( F2 = '' ) then
        F2 := ExpandFileName(S);
    end;
  end;

  LoadFiles( F1, F2 );
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  UpdateHistoryRecent;

  FOptions.WindowPosX.Value := Left;
  FOptions.WindowPosY.Value := Top;
  FOptions.WindowWidth.Value := Width;
  FOptions.WindowHeight.Value := Height;
  FOptions.WindowIsMaximized.Value := WindowState = wsMaximized;

  FOptionsManager.Save;
  FOptionsManager.FreeItems;
  FOptionsManager.Free;

  FDiffDocument.Free;
  FDiffDocument := nil;

  ClearSyntaxDocuments;
  FSyntaxDoc[0].Free;
  FSyntaxDoc[1].Free;
  FSyntaxTokenizer[0].Free;
  FSyntaxTokenizer[1].Free;
  FSyntaxLanguages.Free;

  FOverviewBackground.Free;
end;

function TMainForm.HasDiffDocument: boolean;
begin
  // Note that FormResize may be called after FormDestroy,
  // so we have to make sure that the main document exists too
  Result := (FDiffDocument <> nil) and (FDiffDocument.Count > 0);
end;

procedure TMainForm.ClearSyntaxDocuments;
var
  I, J: integer;
begin
  for J:=0 to 1 do begin
    for I:=1 to FSyntaxDoc[J].Count do
      TSyntaxTokenItem( FSyntaxDoc[J][I-1] ).Free;
    FSyntaxDoc[J].Clear;
  end;
end;

procedure TMainForm.UpdateSyntaxDocuments;
var
  I: integer;
  D: TDiffLine;
  L1, L2: TSyntaxLine;
  F1, F2: TSyntaxLineFlags;
begin
  F1 := [];
  F2 := [];
  for I:=1 to FDiffDocument.Count do begin
    D := FDiffDocument.Line[I-1];
    L1 := nil;
    L2 := nil;
    if( FSyntaxTokenizer[0] <> nil )and( D.Mode in [dmOnlyDoc1,dmBothDocuments,dmCopiedFromDoc1,dmCopiedFromDoc2] ) then begin
      L1 := TSyntaxLine.Create( FSyntaxTokenizer[0] );
      L1.Assign( ExpandTabs(D.Line1,FOptions.TabSize.Value), F1 );
      F1 := L1.Flags;
    end;
    if( FSyntaxTokenizer[1] <> nil )and( D.Mode in [dmOnlyDoc2,dmBothDocuments,dmCopiedFromDoc1,dmCopiedFromDoc2] ) then begin
      L2 := TSyntaxLine.Create( FSyntaxTokenizer[1] );
      L2.Assign( ExpandTabs(D.Line2,FOptions.TabSize.Value), F2 );
      F2 := L2.Flags;
    end;
    FSyntaxDoc[0].Add( L1 );
    FSyntaxDoc[1].Add( L2 );
  end;
end;

// Called when the main document (a TDiffLineList) changes
procedure TMainForm.DiffLineListChanged;
var
  D: TDiffLine;
  I, M: integer;
begin
  FDiffDocument.RebuildDiff;

  // Enable/disable buttons that depend on a non-empty document
  CmdRefresh.Enabled := HasDiffDocument;
  BtnRefresh.Enabled := HasDiffDocument;
  CmdFind.Enabled := HasDiffDocument;
  BtnFind.Enabled := HasDiffDocument;
  CmdFindAgain.Enabled := HasDiffDocument;
  BtnFindAgain.Enabled := HasDiffDocument;
  BtnShowLineNumbers.Enabled := HasDiffDocument;
  BtnSyntaxHilight.Enabled := HasDiffDocument;
  BtnShowSeparators.Enabled := HasDiffDocument;

  if( HasDiffDocument ) then begin
    // Build syntax highlighting
    ClearSyntaxDocuments;
    UpdateSyntaxDocuments;

    // Get the max length of a source line
    M := -1;
    for I:=1 to FDiffDocument.Count do begin
      D := FDiffDocument.Line[I-1];
      if( Length(D.Line1) > M ) then M := Length(D.Line1);
      if( Length(D.Line2) > M ) then M := Length(D.Line2);
    end;

    // Precompute how many characters are needed to print the line numbers
    I := Max( FDiffDocument.Line[FDiffDocument.Count-1].LineNo1,
        FDiffDocument.Line[FDiffDocument.Count-1].LineNo2 );
    FLineNumberLength := Max( MinLineNumberLength, Length(IntToStr(I)) );

    // Enable/disable find difference buttons
    UpdateDifferenceNavigators;

    // Resize controls
    I := HorzScrollBar.Position;
    if( I > M ) then
      I := M;
    HorzScrollBar.SetParams( I, 1, M );

    // Repaint
    FormResize(Self);
    FormPaint(Self);

    // Update status bar
    if( FDiffDocument.DiffCount1+FDiffDocument.DiffCount2 > 0 ) then begin
      StatusBar.SimpleText := ' Files are different (L=+' + IntToStr(FDiffDocument.DiffCount1) + ', R=+' + IntToStr(FDiffDocument.DiffCount2) + ')';
    end
    else begin
      StatusBar.SimpleText := ' No differences found'
    end;
    StatusBar.Hint := StatusBar.SimpleText;
  end;
end;

function TMainForm.GetBlockEnd( L: integer ): integer;
begin
  Result := L;
  repeat
    Inc( Result );
  until( Result >= FDiffDocument.Count )or( FDiffDocument.Mode[Result] <> FDiffDocument.Mode[L] );
  Dec( Result );
end;

function TMainForm.GetBlockStart( L: integer ): integer;
begin
  Result := L;
  repeat
    Dec( Result );
  until( Result < 0 )or( FDiffDocument.Mode[Result] <> FDiffDocument.Mode[L] );
  Inc( Result );
end;

// Changes the status of a difference line or block, can be called with the index
// of any line inside the block
procedure TMainForm.ChangeLineBlock( Line: integer; ModeFrom, ModeTo: TDiffLineMode; Block: boolean );
var
  FirstLineOfBlock: integer;
  FirstLinePastBlock: integer;
  BlockLength: integer;
  D: TDiffLine;
begin
  D := FDiffDocument.Line[ Line ];

  if( D <> nil ) then begin
    Assert( D.Mode = ModeFrom, 'Unexpected line mode for block operation' );
    FirstLineOfBlock := Line;
    FirstLinePastBlock := Line+1;
    if( Block ) then begin
      // In block mode, get the full extent of the block
      while( FirstLineOfBlock > 0 )and( FDiffDocument.Mode[FirstLineOfBlock-1] = ModeFrom ) do
        Dec( FirstLineOfBlock );
      while( FirstLinePastBlock < FDiffDocument.Count )and( FDiffDocument.Mode[FirstLinePastBlock] = ModeFrom ) do
        Inc( FirstLinePastBlock );
    end;
    BlockLength := FirstLinePastBlock - FirstLineOfBlock;

    // ...create entry for undo
    FUndoRedoBuffer.Push( TDiffUndoItem.Create(
        ModeFrom,
        ModeTo,
        FDiffDocument.Line[FirstLineOfBlock].RawLineNo,
        FDiffDocument.Line[FirstLineOfBlock].RawLineNo + BlockLength,
        FFlags) );

    // ...modify block
    while( FirstLineOfBlock < FirstLinePastBlock ) do begin
      FDiffDocument.Line[ FirstLineOfBlock ].Mode := ModeTo;
      Inc( FirstLineOfBlock );
    end;

    // Update modification flags
    if( ModeTo in [dmDeletedFromDoc1,dmCopiedFromDoc2] ) then
      FFlags := FFlags or FlagModifiedFile1;
    if( ModeTo in [dmDeletedFromDoc2,dmCopiedFromDoc1] ) then
      FFlags := FFlags or FlagModifiedFile2;

    ModificationFlagsChanged;
    DiffLineListChanged;
  end;
end;

// Returns the height of one line in pixels
function TMainForm.GetLineHeight: integer;
begin
  Result := Abs( PbxView1.Font.Height ) + 4;
end;

// Returns the number of *fully* visible lines
function TMainForm.GetVisibleLines: integer;
begin
  // Note: take into account the size of the 3D border
  Result := (PbxView1.Height - 4) div GetLineHeight;
end;

function TMainForm.GetDefaultOverviewRect: TRect;
begin
  Result := Rect( 4, 2, PbxOverview.Width-6, PbxOverview.Height-2 );
end;

procedure TMainForm.SetViewPosition( L: integer );
begin
  SetViewPosition( L, false );
end;

procedure TMainForm.SetViewPosition( L: integer; DontNotify: boolean );
var
  E: procedure (Sender: TObject) of object;
begin
  if( L < 0 ) then
    L := 0;
  if( L > VertScrollBar.Max ) then
    L := VertScrollBar.Max;
  E := VertScrollBar.OnChange;
  if( DontNotify ) then begin
    VertScrollBar.OnChange := nil;
  end;
  VertScrollBar.Position := L;
  VertScrollBar.OnChange := E;
end;

procedure TMainForm.DrawOverviewBackground( C: TCanvas; var R: TRect );
const
  Darkness = $50;
var
  L1, L2: integer;
  M, M1, M2: TDiffLineMode;
  R1, R2: TRect;
  I: integer;
  W: integer;
procedure DrawBlock( Rect: TRect; Color: TColor );
  var
    R, G, B: byte;
  begin
    if( FOverviewDiffsAreSparse ) then begin
      C.Brush.Color := Color;
      if( Rect.Bottom < 3+Rect.Top ) then
        Rect.Bottom := 3+Rect.Top;
      C.Rectangle( Rect );
    end
    else begin
      // Make color darker to increase contrast
      R := GetRValue( Color ); if( R > Darkness ) then Dec(R,Darkness) else R := 0;
      G := GetGValue( Color ); if( G > Darkness ) then Dec(G,Darkness) else G := 0;
      B := GetBValue( Color ); if( B > Darkness ) then Dec(B,Darkness) else B := 0;
      C.Brush.Color := RGB(R,G,B);
      if( Rect.Bottom < 1+Rect.Top ) then
        Rect.Bottom := 1+Rect.Top;
      // Do not draw any border around this rectangle!
      C.FillRect( Rect );
    end;
  end;
begin
  if( not HasDiffDocument ) then
    Exit;

  C.Pen.Color := clBlack;
  C.Brush.Color := clBtnFace;
  C.FillRect( R );
  W := ((R.Right - R.Left) - 2) div 2;
  R1 := Rect( R.Left, 0, R.Left + W, 0 );
  R2 := Rect( R.Right-W, 0, R.Right, 0 );

  L1 := 0;
  L2 := 0;

  M1 := dmNone;
  M2 := dmNone;
  for I:=0 to FDiffDocument.Count do begin
    M := dmNone;
    if( I < FDiffDocument.Count ) then
      M := FDiffDocument.Mode[ I ];

    // Track lines that are only in file one
    if( M = dmOnlyDoc1 ) then begin
      if( M1 <> dmOnlyDoc1 ) then begin
        L1 := I;
        M1 := dmOnlyDoc1;
      end
    end
    else if( M1 = dmOnlyDoc1 ) then begin
      // This is the end of a "file one only" block
      R1.Top := R.Top + L1 * (R.Bottom-R.Top) div FDiffDocument.Count;
      R1.Bottom := R.Top + I * (R.Bottom-R.Top) div FDiffDocument.Count;
      DrawBlock( R1, FOptions.ColorOnlyDoc1.Value );
      M1 := dmNone;
    end;

    // Track lines that are only in file two
    if( M = dmOnlyDoc2 ) then begin
      if( M2 <> dmOnlyDoc2 ) then begin
        L2 := I;
        M2 := dmOnlyDoc2;
      end
    end
    else if( M2 = dmOnlyDoc2 ) then begin
      // This is the end of a "file two only" block
      R2.Top := R.Top + L2 * (R.Bottom-R.Top) div FDiffDocument.Count;
      R2.Bottom := R.Top + I * (R.Bottom-R.Top) div FDiffDocument.Count;
      DrawBlock( R2, FOptions.ColorOnlyDoc2.Value );
      M2 := dmNone;
    end;
  end;
end;

function TMainForm.GetOverviewThumbRect: TRect;
var
  R1: TRect;
begin
  Assert( FDiffDocument.Count > 0 );
  R1 := GetDefaultOverviewRect;
  Result := Rect( R1.Left,
    R1.Top + VertScrollBar.Position * ((R1.Bottom-R1.Top)) div FDiffDocument.Count,
    R1.Right,
    0 );
  Result.Bottom := Min(R1.Bottom,2 + Result.Top + ((GetVisibleLines) * (R1.Bottom-R1.Top)) div FDiffDocument.Count);
end;

procedure TMainForm.FormPaint(Sender: TObject);
const
  TextLeftMargin = 1;
var
  I: integer;
  Y: integer;
  H: integer;
  N: integer;
  S: string;
  L: TDiffLine;
  C1, C2: TCanvas;
  R1, R2: TRect;
  B1, B2: TColor;
  LineNumberWidth: integer;
procedure DrawLineNumber( Canvas: TCanvas; LineNo: integer; var ClipRect: TRect );
  var
    R: TRect;
  begin
    if( LineNumberWidth > 0 ) then begin
      R := Rect( ClipRect.Left, ClipRect.Top, ClipRect.Left+LineNumberWidth-1, ClipRect.Bottom );
      // Draw the line number
      Canvas.Font.Style := [];
      if( B1 = clHighlight ) then
        Canvas.Brush.Color := clHighlight
      else
        Canvas.Brush.Color := FOptions.ColorLineNumber.Value;
      if( LineNo >= 0 ) then
        Canvas.TextRect( R, R.Left+1, R.Top, Format('%*d', [FLineNumberLength, LineNo+1]) )
      else
        Canvas.FillRect( R );
      // Draw a gray border around the line number rect
      R.Left := R.Right;
      Inc( R.Right );
      Canvas.Brush.Color := clGrayText;
      Canvas.FillRect( R );
      // Substract the rectangle with just painted from the clip area
      Inc( ClipRect.Left, LineNumberWidth );
    end;
  end;
procedure DrawLineWithHighlight( Canvas: TCanvas; SL: TSyntaxLine; var ClipRect: TRect );
  var
    T: TSyntaxTokenItem;
    S: string;
    I: integer;
    X: integer;
  begin
    // Draw an empty background first
    Canvas.FillRect( ClipRect );
    // Start drawing the line tokens
    if( SL <> nil ) then begin
      X := ClipRect.Left + TextLeftMargin;
      for I:=1 to SL.TokenCount do begin
        T := SL.Tokens[I-1];
        if( T.Pos + T.Len >= HorzScrollBar.Position ) then begin
          // This token is visible, print it
          if( T.Pos < HorzScrollBar.Position ) then
            S := Copy( SL.Line, HorzScrollBar.Position, T.Len-HorzScrollBar.Position+T.Pos ) // Token partially visible
          else
            S := Copy( SL.Line, T.Pos, T.Len ); // Token fully visible
          // Set the display attributes
          case T.Kind of
            tkComment: Canvas.Font.Style := [fsItalic];
            tkIdentifier: Canvas.Font.Style := [fsBold];
            else Canvas.Font.Style := [];
          end;
          // Print the token
          Canvas.TextOut( X, ClipRect.Top, S );
          Inc( X, Canvas.TextWidth( S ) );
        end;
      end; // for
    end; // if SL not nil
  end;
procedure DrawLine( B1, B2: integer );
  begin
    // Left file
    if( B1 <> 0 ) then begin
      DrawLineNumber( C1, L.LineNo1, R1 );
      C1.Brush.Color := B1;
      if( FOptions.SyntaxHighlight.Value )and( FSyntaxTokenizer[0] <> nil ) then begin
        DrawLineWithHighlight( C1, TSyntaxLine(FSyntaxDoc[0][N]), R1 );
      end
      else begin
        S := ExpandTabs( L.Line1, FOptions.TabSize.Value );
        if( HorzScrollBar.Position > 1 ) then
          S := Copy( S, HorzScrollBar.Position, Length(S) );
        C1.TextRect( R1, R1.Left + TextLeftMargin, R1.Top, S );
      end;
    end
    else begin
      C1.Brush.Color := FOptions.ColorMissingLine.Value;
      C1.FillRect( R1 );
    end;

    // Right file
    if( B2 <> 0 ) then begin
      DrawLineNumber( C2, L.LineNo2, R2 );
      C2.Brush.Color := B2;
      if( FOptions.SyntaxHighlight.Value )and( FSyntaxTokenizer[1] <> nil ) then begin
        DrawLineWithHighlight( C2, TSyntaxLine(FSyntaxDoc[1][N]), R2 );
      end
      else begin
        S := ExpandTabs( L.Line2, FOptions.TabSize.Value );
        if( HorzScrollBar.Position > 1 ) then
          S := Copy( S, HorzScrollBar.Position, Length(S) );
        C2.TextRect( R2, R2.Left + TextLeftMargin, R2.Top, S );
      end;
    end
    else begin
      C2.Brush.Color := FOptions.ColorMissingLine.Value;
      C2.FillRect( R2 );
    end;
  end;
begin
  // Setup paint boxes
  C1 := PbxView1.Canvas;
  C1.Font.Assign( PbxView1.Font );
  C2 := PbxView2.Canvas;
  C2.Font.Assign( PbxView1.Font );

  // Initialize variables
  H := GetLineHeight;

  LineNumberWidth := 0;
  if( FOptions.ShowLineNumbers.Value ) then begin
    // This assumes line numbers are shown with a fixed pitch font
    LineNumberWidth := 4 + PbxView1.Canvas.TextWidth( StringOfChar('9',FLineNumberLength) );
  end;

  // Render left and right views
  for I:=0 to GetVisibleLines do begin // Note: there is one more line *intentionally*
    N := I + VertScrollBar.Position;
    L := FDiffDocument.Line[ N ];
    Y := 2 + I*H;
    R1 := Rect( 2, Y, PbxView1.Width-2, Y+H );
    R2 := Rect( 2, Y, PbxView2.Width-2, Y+H );
    B1 := 0;
    B2 := 0;
    if( L <> nil ) then begin
      // Draw a separator line if needed
      if( FOptions.ShowBlockSeparators.Value )and( I > 0 )and( FDiffDocument.Mode[ N-1 ] <> L.Mode ) then begin
        C1.Brush.Color := clBlack;
        C1.FillRect( Rect( 0, Y-1, PbxView1.Width, Y ) );
        C2.Brush.Color := clBlack;
        C2.FillRect( Rect( 0, Y-1, PbxView2.Width, Y ) );
      end;

      // Draw text with proper color and style
      if( N = FCurrentLine ) then begin
        B1 := clHighlight;
        B2 := clHighlight;
        C1.Font.Color := clHighlightText;
        C2.Font.Color := clHighlightText;
      end
      else begin
        C1.Font.Color := clBlack;
        C2.Font.Color := clBlack;
        case L.Mode of
          dmOnlyDoc1: begin
            B1 := FOptions.ColorOnlyDoc1.Value;
            B2 := 0;
          end;
          dmOnlyDoc2: begin
            B1 := 0;
            B2 :=FOptions.ColorOnlyDoc2.Value;
          end;
          dmBothDocuments: begin
            B1 := FOptions.ColorDefault.Value;
            B2 := B1;
          end;
          dmCopiedFromDoc1,
          dmCopiedFromDoc2: begin
            B1 := FOptions.ColorChangedLine.Value;
            B2 := B1;
          end;
        end;
      end;
    end;
    DrawLine( B1, B2 );
  end;

  // Draw the 3D border around left and right views
  R1 := Rect( 0, 0, PbxView1.Width, PbxView1.Height );
  DrawEdge( C1.Handle, R1, EDGE_SUNKEN, BF_RECT );
  R2 := Rect( 0, 0, PbxView2.Width, PbxView2.Height );
  DrawEdge( C2.Handle, R2, EDGE_SUNKEN, BF_RECT );

  // Render overview
  if( HasDiffDocument ) then begin
    R1 := GetDefaultOverviewRect;
    R2 := GetOverviewThumbRect;
    with PbxOverview.Canvas do begin
      Draw( R1.Left, R1.Top, FOverviewBackground );
      Brush.Color := clBlack;
      FrameRect( R2 );
      InflateRect( R2, -1, -1 );
      FrameRect( R2 );
    end;
  end;
end;

procedure TMainForm.FormResize(Sender: TObject);
var
  M: integer;
  N: integer;
  R: TRect;
begin
  // Resize file panels
  N := ClientWidth - PbxOverview.Width - Splitter.Width - VertScrollBar.Width;
  N := N * FSizeView1 div 100;
  if( Abs(PbxView1.Width-N) > 10 ) then begin
    // Resize only if the change is noticeable, otherwise it gets
    // quite annoying!
    PbxView1.Width := N;
  end;

  HorzScrollBar.Left := PbxView1.Left + 1;
  HorzScrollBar.Width := PbxView1.Width + Splitter.Width + PbxView2.Width - 2;

  // Resize status bar panels
  StatusBar.Panels[0].Width := Width * 3 div 4;
  StatusBar.SimpleText := StatusBar.Hint;

  if( HasDiffDocument ) then begin
    // Update the vertical scroll bar
    N := GetVisibleLines;
    M := 0;
    if( N < FDiffDocument.Count ) then
      M := FDiffDocument.Count - N;
    if( VertScrollBar.Position > M ) then
      VertScrollBar.Position := M;
    VertScrollBar.Max := M;
    VertScrollBar.LargeChange := N;

    // Draw the overview background (cached into a bitmap)
    R := GetDefaultOverviewRect;
    OffsetRect( R, -R.Left, -R.Top );
    with FOverviewBackground do begin
      Width := R.Right - R.Left;
      Height := R.Bottom - R.Top;
      DrawOverviewBackground( Canvas, R );
    end;
  end;
end;

procedure TMainForm.ScrollBarChange(Sender: TObject);
begin
  FormPaint(Self);
  UpdateDifferenceNavigators;
  StatusBar.SimpleText := StatusBar.Hint;
end;

// Updates the GUI when the modification flags have changed
procedure TMainForm.ModificationFlagsChanged;
var
  F: boolean;
begin
  // Left file
  F := FFlags and FlagModifiedFile1 <> 0;
  BtnSaveLeft.Enabled := F;
  CmdSaveLeftFile.Enabled := F;
  PopSaveLeftFile.Enabled := F;
  CmdSaveLeftFileAs.Enabled := F;
  PopSaveLeftFileAs.Enabled := F;

  // Right file
  F := FFlags and FlagModifiedFile2 <> 0;
  BtnSaveRight.Enabled := F;
  CmdSaveRightFile.Enabled := F;
  PopSaveRightFile.Enabled := F;
  CmdSaveRightFileAs.Enabled := F;
  PopSaveRightFileAs.Enabled := F;

  BtnSaveAll.Enabled := BtnSaveLeft.Enabled or BtnSaveRight.Enabled;
  CmdSaveAll.Enabled := BtnSaveAll.Enabled;

  // Undo/redo buttons
  BtnRedo.Enabled := FUndoRedoBuffer.RedoCount > 0;
  CmdRedo.Enabled := FUndoRedoBuffer.RedoCount > 0;
  BtnUndo.Enabled := FUndoRedoBuffer.UndoCount > 0;
  CmdUndo.Enabled := FUndoRedoBuffer.UndoCount > 0;
end;

function TMainForm.AllowCloseCurrentDocument: boolean;
  function AskSaveFile( Filename: string; Mode: TDiffLineMode ): boolean;
  var
    R: integer;
  begin
    Result := true;
    R := MessageDlg( 'File ''' + Filename + ''' has been modified, do you want to save it?',
      mtWarning, [mbYes,mbNo,mbCancel], 0 );
    case R of
      mrYes: Result := SaveFile(Filename, Mode);
      mrNo: ; // Nothing to do then
      else begin // Cancel
        Result := false;
        Exit;
      end
    end;
  end;
begin
  Result := true;

  if( FFlags and FlagModifiedFile1 <> 0 ) then begin
    Result := AskSaveFile( FFilename1, dmOnlyDoc1 );
  end;

  if( Result )and( FFlags and FlagModifiedFile2 <> 0 ) then begin
    Result := AskSaveFile( FFilename2, dmOnlyDoc2 );
  end;
end;

procedure TMainForm.UpdateRecentFilesMenu;
const
  MenuStartPos : integer = 7;
var
  I: integer;
  M: TMenuItem;
  P: integer;
  S: string;
  T: string;
begin
  // First, remove all "recent files" items from the menu
  // (they start from the 6th position and finish in the 3rd from
  // last position included)
  for I:=CmdFile.Count-3 downto MenuStartPos do begin
    M := CmdFile.Items[I];
    CmdFile.Delete( I );
    M.Free;
  end;

  // Exit if nothing to do
  if( FOptions.HistoryRecent.Count <= 0 ) then
    Exit;

  // Insert the updated item list
  for I:=FOptions.HistoryRecent.Count-1 downto 0 do begin
    M := TMenuItem.Create( Self );
    S := '&' + IntToStr(I+1) + ' ' + FOptions.HistoryRecent.Strings[I];
    P := Pos( '|', S );
    Assert( P > 0 );
    T := Copy( S, P+1, Length(S) );
    S := Copy( S, 1, P-1 );
    M.Caption := S;
    M.Tag := I;
    M.OnClick := CmdRecentFilesClick;
    P := Pos( '|', T );
    Assert( P > 0 );
    Delete( T, P, 1 );
    Insert( ' / ', T, P );
    M.Hint := T;
    CmdFile.Insert( MenuStartPos, M );
  end;

  // Insert separator
  M := TMenuItem.Create( Self );
  M.Caption := '-';
  CmdFile.Insert( MenuStartPos, M );
end;

procedure TMainForm.UpdateHistoryRecent;
var
  N1, N2, S: string;
  I: integer;
begin
  N1 := ExtractFileName( FFileName1 );
  N2 := ExtractFileName( FFileName2 );
  if( N1 <> '' )and( N2 <> '' ) then begin
    S := N1 + '/' + N2 + '|' + FFileName1 + '|' + FFileName2;
    I := FOptions.HistoryRecent.Strings.IndexOf(S);
    if( I >= 0 ) then begin // Already in the list, move to first position
      FOptions.HistoryRecent.Strings.Move( I, 0 );
    end
    else begin // Add to list
      FOptions.HistoryRecent.Strings.Insert( 0, S );
      while( FOptions.HistoryRecent.Count > MaxRecentFilesCount ) do begin
        FOptions.HistoryRecent.Strings.Delete( FOptions.HistoryRecent.Count-1 );
      end;
    end;
  end;
end;

procedure TMainForm.UpdateFormCaption;
const
  MaxFileCaptionLen = 60;
var
  C: string;
  A1, A2: boolean;
begin
  C := FDefaultCaption;
  A1 := FFileName1 <> '';
  A2 := FFileName2 <> '';
  if( A1 )or( A2 ) then begin
    C := C + ' - ';
    if( A1 ) then  C := C + ShortenFilePath(FFileName1,MaxFileCaptionLen);
    if( A1 and A2 ) then C := C + ' / ';
    if( A2 ) then  C := C + ShortenFilePath(FFileName2,MaxFileCaptionLen);
  end;
  Caption := C;
end;

procedure LoadFile( FileName: string; SL: TStringList );
begin
  if( FileName <> '' ) then begin
    SL.LoadFromFile( FileName );
  end
  else begin
    SL.Clear;
  end;
end;

// Loads and differences two files
function TMainForm.LoadFiles( F1, F2: string ): boolean;
var
  SL1, SL2: TStringList;
procedure SetupSyntaxTokenizer( const F: string; var T: TSyntaxTokenizer );
  var
    E: string;
    I: integer;
    O: TLanguageOptions;
  begin
    O := nil;
    E := LowerCase( ' ' + ExtractFileExt( F ) + ' ' );
    for I:=1 to FSyntaxLanguages.Count do begin
      O := TLanguageOptions( FSyntaxLanguages[I-1] );
      if( AnsiPos( E, O.Extensions.Value ) > 0 ) then
        break;
      O := nil;
    end;
    if( O = nil ) then begin
      T.Free;
      T := nil;
    end
    else if( T = nil )or( T.Name <> O.Name ) then begin
      if( T = nil ) then begin
        T := TSyntaxTokenizer.Create( O.Name );
      end;
      T.Identifiers.Clear;
      T.AddIdentifiers( O.Identifiers.Value );
      T.IgnoreCase := O.IgnoreCase.Value;
      T.QuoteChars := O.QuoteChars.Value;
      T.QuoteEscape := PChar(O.QuoteChars.Value)^;
      T.LineComment := O.LineComment.Value;
      T.CommentBegin := O.CommentBegin.Value;
      T.CommentEnd := O.CommentEnd.Value;
    end;
  end;
begin
  Result := false;

  // Set default values
  if( F1 = '' ) then F1 := FFileName1;
  if( F2 = '' ) then F2 := FFileName2;

  if( F1 = '' )and( F2 = '' ) then // Nothing to do!
    Exit;

  if( not AllowCloseCurrentDocument ) then
    Exit;

  try
    Screen.Cursor := crHourglass;

    SL1 := nil;
    SL2 := nil;
    try
      SL1 := TStringList.Create;
      SL2 := TStringList.Create;
      try
        LoadFile( F1, SL1 );
        LoadFile( F2, SL2 );
        FDiffDocument.RawDoc1.Assign( SL1 );
        FDiffDocument.RawDoc2.Assign( SL2 );
        Result := true;
      except
        on E: Exception do
          MessageDlg( E.Message, mtError, [mbOk], 0 );
      else
          MessageDlg( 'Cannot load file(s).', mtError, [mbOk], 0 );
      end;
    finally
      SL1.Free;
      SL2.Free;
    end;

    if( Result ) then begin
      FDiffDocument.DiffEngine.FIgnoreBlanks := FOptions.IgnoreBlanks.Value;
      FDiffDocument.DiffEngine.FIgnoreCase := FOptions.IgnoreCase.Value;
      FDiffDocument.RebuildAll;

      UpdateHistoryRecent;
      UpdateRecentFilesMenu;

      FFileName1 := F1;
      FFileName2 := F2;
      UpdateFormCaption;
      FFlags := 0; // No file modified yet
      FUndoRedoBuffer.Clear;

      // Setup tokenizers for syntax highlighting
      SetupSyntaxTokenizer( F1, FSyntaxTokenizer[0] );
      SetupSyntaxTokenizer( F2, FSyntaxTokenizer[1] );

      SetViewPosition( 0, true ); // Don't paint yet!
      ModificationFlagsChanged;
      DiffLineListChanged;
    end;
  finally
    Screen.Cursor := crDefault;
  end;
end;

procedure TMainForm.CmdOpenClick(Sender: TObject);
const
  Initialize: boolean = true;
begin
  if( Initialize ) then begin
    Initialize := false;
    OpenForm.CboLeft.Items.Assign( FOptions.HistoryFile1.Strings );
    OpenForm.CboRight.Items.Assign( FOptions.HistoryFile2.Strings );
    OpenForm.SetFileFilters( FOptions.OpenFileFilters.Value );
  end;

  OpenForm.LeftFile := FFilename1;
  OpenForm.RightFile := FFilename2;

  OpenForm.SetFocusToRightFile( Sender = PopOpenRightFile );

  if( OpenForm.ShowModal = mrOk ) then begin
    if( LoadFiles( OpenForm.LeftFile, OpenForm.RightFile ) ) then begin
      FOptions.HistoryFile1.Strings.Assign( OpenForm.CboLeft.Items );
      FOptions.HistoryFile2.Strings.Assign( OpenForm.CboRight.Items );
    end;
  end;
end;

procedure TMainForm.CmdExitClick(Sender: TObject);
begin
  Close;
end;

procedure TMainForm.PbxView1MouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  L: integer;
  P: TPoint;
begin
  // Get the index of the line under the cursor
  L := (Y - 2) div GetLineHeight + VertScrollBar.Position;
  if( L >= 0 )and( L < FDiffDocument.Count ) then case Button of
    mbRight: begin
      PopCopySingleLineToRight.Enabled := FDiffDocument.Mode[L] = dmOnlyDoc1;
      PopCopyToRight.Enabled := FDiffDocument.Mode[L] = dmOnlyDoc1;
      PopDeleteLineLeft.Enabled := FDiffDocument.Mode[L] = dmOnlyDoc1;
      PopDeleteBlockLeft.Enabled := FDiffDocument.Mode[L] = dmOnlyDoc1;
      PopupLeft.Tag := L; // Save selected line into popup menu tag
      P := ClientToScreen( Point( PbxView1.Left+X, PbxView1.Top+Y ) );
      PopupLeft.Popup( P.X, P.Y );
    end;
    mbLeft: if( Shift = [] ) then begin
    end;
  end;
end;

procedure TMainForm.PbxView2MouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  L: integer;
  P: TPoint;
begin
  // Get the index of the line under the cursor
  L := (Y - 2) div GetLineHeight + VertScrollBar.Position;
  if( L >= 0 )and( L < FDiffDocument.Count ) then case Button of
    mbRight: begin
      PopCopySingleLineToLeft.Enabled := FDiffDocument.Mode[L] = dmOnlyDoc2;
      PopCopyLinesToLeft.Enabled := FDiffDocument.Mode[L] = dmOnlyDoc2;
      PopDeleteLineRight.Enabled := FDiffDocument.Mode[L] = dmOnlyDoc2;
      PopDeleteBlockRight.Enabled := FDiffDocument.Mode[L] = dmOnlyDoc2;
      PopupRight.Tag := L; // Save selected line into popup menu tag
      P := ClientToScreen( Point( PbxView2.Left+X, PbxView2.Top+Y ) );
      PopupRight.Popup( P.X, P.Y );
    end;
    mbLeft: if( Shift = [] ) then begin
    end;
  end;
end;

procedure TMainForm.ChangeBlock( L: integer; Action: TChangeBlockAction );
begin
  case Action of
    baCopyToDoc2,
    baCopySingleToDoc2:
      ChangeLineBlock( L, dmOnlyDoc1, dmCopiedFromDoc1, Action = baCopyToDoc2 );
    baCopyToDoc1,
    baCopySingleToDoc1:
      ChangeLineBlock( L, dmOnlyDoc2, dmCopiedFromDoc2, Action = baCopyToDoc1 );
    baDeleteFromDoc1,
    baDeleteSingleFromDoc1:
      ChangeLineBlock( L, dmOnlyDoc1, dmDeletedFromDoc1, Action = baDeleteFromDoc1 );
    baDeleteFromDoc2,
    baDeleteSingleFromDoc2:
      ChangeLineBlock( L, dmOnlyDoc2, dmDeletedFromDoc2, Action = baDeleteFromDoc2 );
  end;
end;

procedure TMainForm.PopCopyToRightClick(Sender: TObject);
begin
  ChangeBlock( PopupLeft.Tag, baCopyToDoc2 );
end;

procedure TMainForm.PopCopySingleLineToRightClick(Sender: TObject);
begin
  ChangeBlock( PopupLeft.Tag, baCopySingleToDoc2 );
end;

procedure TMainForm.PopDeleteBlockLeftClick(Sender: TObject);
begin
  ChangeBlock( PopupLeft.Tag, baDeleteFromDoc1 );
end;

procedure TMainForm.PopDeleteLineLeftClick(Sender: TObject);
begin
  ChangeBlock( PopupLeft.Tag, baDeleteSingleFromDoc1 );
end;

procedure TMainForm.PopCopyLinesToLeftClick(Sender: TObject);
begin
  ChangeBlock( PopupRight.Tag, baCopyToDoc1 );
end;

procedure TMainForm.PopCopySingleLineToLeftClick(Sender: TObject);
begin
  ChangeBlock( PopupRight.Tag, baCopySingleToDoc1 );
end;

procedure TMainForm.PopDeleteBlockRightClick(Sender: TObject);
begin
  ChangeBlock( PopupRight.Tag, baDeleteFromDoc2 );
end;

procedure TMainForm.PopDeleteLineRightClick(Sender: TObject);
begin
  ChangeBlock( PopupRight.Tag, baDeleteSingleFromDoc2 );
end;

procedure TMainForm.CmdAboutClick(Sender: TObject);
begin
  AboutForm := TAboutForm.Create(Self);
  AboutForm.ShowModal;
  AboutForm.Free;
end;

procedure TMainForm.FormMouseWheel(Sender: TObject; Shift: TShiftState;
  WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
begin
  SetViewPosition( VertScrollBar.Position + WheelDelta div WHEEL_DELTA );
  Handled := true;
end;

function TMainForm.GetLineFromOverview( Y: integer ): integer;
var
  R: TRect;
  L: integer;
begin
  Result := -1;
  if( HasDiffDocument ) then begin
    R := GetDefaultOverviewRect;
    if( Y < R.Top ) then
      Y := R.Top;
    L := (Y - R.Top) * FDiffDocument.Count div (R.Bottom-R.Top);
    if( L >= 0 )and( L < FDiffDocument.Count ) then begin
      Result := L;
    end;
  end;
end;

const
  FOvMouseDown: boolean = false;
var
  FOvDragging: boolean;
  FOvMousePos: TPoint;
  FOvDragLine: integer;

procedure TMainForm.PbxOverviewMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  L: integer;
begin
  if( HasDiffDocument )and( FOvMouseDown )and( not FOvDragging ) then begin
    L := GetLineFromOverview( Y );
    if( L >= 0 ) then begin
      Dec( L, GetVisibleLines div 2 );
      SetViewPosition( L );
    end;
  end;
  FOvMouseDown := false;
end;

procedure TMainForm.PbxOverviewMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  R: TRect;
begin
  if( Button = mbRight ) then begin
    FOverviewDiffsAreSparse := not FOverviewDiffsAreSparse;
    R := GetDefaultOverviewRect;
    OffsetRect( R, -R.Left, -R.Top );
    DrawOverviewBackground( FOverviewBackground.Canvas, R );
    FormPaint(Self);
  end
  else if( Button = mbLeft ) then begin
    FOvMouseDown := true;
    FOvMousePos := Point( X, Y );
    R := GetOverviewThumbRect;
    FOvDragging := PtInRect( R, FOvMousePos );
    FOvDragLine := VertScrollBar.Position;
  end;
end;

procedure TMainForm.PbxOverviewMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
var
  L: integer;
begin
  if( not FOvMouseDown ) then
    Exit;
  if( FOvDragging ) then begin
    if( Y > FOvMousePos.Y ) then
      L := FOvDragLine + GetLineFromOverview( Y-FOvMousePos.Y )
    else
      L := FOvDragLine - GetLineFromOverview( FOvMousePos.Y-Y );
    SetViewPosition( L );
  end;
end;

procedure TMainForm.SplitterMoved(Sender: TObject);
var
  N: integer;
begin
  N := ClientWidth - PbxOverview.Width - Splitter.Width - VertScrollBar.Width;
  FSizeView1 := PbxView1.Width * 100 div N;
end;

procedure TMainForm.ExecuteUndoRedo( Item: TDiffUndoItem; NewMode: TDiffLineMode );
var
  I: integer;
  F: integer;
begin
  for I:=Item.LineFrom to Item.LineTo-1 do
    FDiffDocument.RawLine[I].Mode := NewMode;

  if( Item.Flags >= 0 ) then begin
    F := FFlags;
    FFlags := Item.Flags;
    Item.Flags := F;
    ModificationFlagsChanged;
    DiffLineListChanged;
  end;
end;

procedure TMainForm.OnRedoEvent( Sender: TObject; Item: TUndoRedoItem );
begin
  if( Item is TEditDiffDocUndoItem ) then begin
    ExecuteUndoEditDoc( TEditDiffDocUndoItem(Item), true );
  end
  else begin
    ExecuteUndoRedo( TDiffUndoItem(Item), TDiffUndoItem(Item).NewMode );
  end;
end;

procedure TMainForm.ExecuteUndoEditDoc( Item: TEditDiffDocUndoItem; IsRedo: boolean );
var
  F: integer;
begin
  if( Item.Mode = dmOnlyDoc1 ) then
    FDiffDocument.DiffEngine.Merge( FDiffDocument.RawDoc1, Item.Diffs, IsRedo )
  else
    FDiffDocument.DiffEngine.Merge( FDiffDocument.RawDoc2, Item.Diffs, IsRedo );
  FDiffDocument.RebuildAll;
  F := FFlags;
  FFlags := Item.Flags;
  Item.Flags := F;
  ModificationFlagsChanged;
  DiffLineListChanged;
end;

procedure TMainForm.OnUndoEvent( Sender: TObject; Item: TUndoRedoItem );
begin
  if( Item is TEditDiffDocUndoItem ) then begin
    ExecuteUndoEditDoc( TEditDiffDocUndoItem(Item), false );
  end
  else begin
    ExecuteUndoRedo( TDiffUndoItem(Item), TDiffUndoItem(Item).OldMode );
  end;
end;

procedure TMainForm.ActionUndoExecute(Sender: TObject);
begin
  FUndoRedoBuffer.Undo;
end;

procedure TMainForm.ActionRedoExecute(Sender: TObject);
begin
  FUndoRedoBuffer.Redo;
end;

// Handles file dragged from another application
procedure TMainForm.OnDropFiles( var M: TMessage );
var
  Drop: THandle;
  Name: array[ 1..MAX_PATH ] of char;
  N: integer;
  F1, F2: string;
  DW: DWORD;
  P: TPoint;
begin
  Drop := M.WParam;
  N := DragQueryFile( Drop, $FFFFFFFF, nil, 0 );
  if( N > 0 ) then begin
    F1 := '';
    F2 := '';
    // Get name of first file
    DragQueryFile( Drop, 0, @Name, sizeof(Name) );
    if( N = 1 ) then begin
      // Only one file was dropped, need to use mouse position to
      // select left or right window
      DW := GetMessagePos;
      P.X := LOWORD(DW);
      P.Y := HIWORD(DW);
      P := ScreenToClient( P );
      if( PtInRect( PbxView1.BoundsRect, P ) ) then
        F1 := StrPas(@Name)
      else
        F2 := StrPas(@Name);
    end
    else begin
      // Two or more files, use first two names
      F1 := StrPas(@Name);
      DragQueryFile( Drop, 1, @Name, sizeof(Name) );
      F2 := StrPas(@Name);
    end;
    LoadFiles( F1, F2 );
  end;
  DragFinish( Drop );
  M.Result := 0;
end;

const
  FlagDontShowErrorMessage = $40000000;

function TMainForm.RewriteFile( Filename: string; var F: TextFile ): integer;
var
  A: integer;
  R: integer;
  Done: boolean;
begin
  AssignFile( F, Filename );

  repeat
    Done := true;
{$I-}
    Rewrite( F );
    Result := IOResult;
{$I+}
    if( Result <> 0 ) then begin
      // Failed to open the file, can we do somethihg about it?

      // Access denied: is the file read-only?
      if( Result = 5 ) then begin
        A := FileGetAttr(Filename);
        if( A and faReadOnly <> 0 ) then begin
          R := MessageDlg( 'File ''' + Filename + ''' is read-only. Do you want to override the read-only attribute and save it anyway?',
            mtWarning, [mbYes,mbNo,mbCancel], 0 );
          if( R = mrYes ) then begin
            // Try to remove the read-only attribute
            if( FileSetAttr( Filename, A and (not faReadOnly) ) <> 0 ) then
              MessageDlg( 'The read-only attribute could not be removed!', mtError, [mbOk], 0 )
            else
              Done := false; // Try again!
          end;
          // Since we already displayed an error message, make sure
          // the caller doesn't show it again
          Result := Result or FlagDontShowErrorMessage;
        end
      end;
    end;
  until( Done );
end;

// Saves either the "left" or "right" part of the document (according
// to the specified mode) to file
function TMainForm.SaveFile( Filename: string; Mode: TDiffLineMode ): boolean;
type
  TErrorRec = record
    Code: integer;
    Msg: string;
  end;
const
  ErrorMap: array[0..4] of TErrorRec = (
    (Code: 2; Msg: 'the file was not found' ),
    (Code: 3; Msg: 'the path name is invalid' ),
    (Code: 4; Msg: 'there are too many open files at this time' ),
    (Code: 5; Msg: 'write access to the file was denied by the operating system' ),
    (Code: 101; Msg: 'there is not enough space on the disk' ));
var
  B: string;
  E: string;
  F: TextFile;
  I: integer;
  R: integer;
  L: TDiffLine;
  M: integer;
begin
  Result := false;

  // Backup file if needed
  B := '';
  E := ExtractFileExt( Filename );
  case FOptions.BackupMode.Value of
    0: B := Filename + '.bak';
    1: B := ChangeFileExt( Filename, '.bak'+E );
    2: B := ChangeFileExt( Filename, '.$'+Copy(E,2,Length(E)-2) );
    3: B := ChangeFileExt( Filename, '.~'+Copy(E,2,Length(E)) );
  end;
  if( B <> '' )and( FileExists(Filename) ) then begin
    if( not CopyFile( PChar(Filename), PChar(B), false ) ) then begin
      MessageDlg( 'Cannot create backup file ''' + B + '''. Save operation aborted.',
        mtError, [mbOk], 0 );
      Exit;
    end;
    FileSetAttr( B, faArchive );
  end;

  // Create the output file
  R := RewriteFile( Filename, F );

  if( R <> 0 ) then begin
    if( R and FlagDontShowErrorMessage = 0 ) then begin
      R := R and (not FlagDontShowErrorMessage);
      E := '.'; // Default is no additional info for the error code
      for I:=Low(ErrorMap) to High(ErrorMap) do begin
        if( ErrorMap[I].Code = R ) then begin
          E := ' because ' + ErrorMap[I].Msg + '.';
          break;
        end;
      end;
      MessageDlg( 'Cannot save file ''' + Filename + '''' + E, mtError, [mbOk], 0 );
    end;
    Exit;
  end;

  try
    for I:=1 to FDiffDocument.Count do begin
      L := FDiffDocument.Line[I-1];
      if( L.Mode = Mode )or( L.Mode in [dmBothDocuments,dmCopiedFromDoc1,dmCopiedFromDoc2] ) then begin
        // We output the lines that belong to both files, plus those
        // that belong only to the specified file (dmOnlyDoc1 or dmOnlyDoc2)
        if( Mode = dmOnlyDoc1 ) then begin
          WriteLn( F, L.Line1 )
        end
        else begin
          WriteLn( F, L.Line2 );
        end;
      end;
    end;
  except
    on Ex: Exception do begin
      MessageDlg( 'Cannot write file. ' + Ex.Message + '.', mtError, [mbOk], 0 );
      CloseFile( F );
      Exit;
    end;
  end;

  CloseFile( F );

  // If we managed to save a file without errors, we must now update
  // the undo/redo buffers, because from now on every undo/redo action
  // will cause the corresponding document to change.
  M := FlagModifiedFile2;
  if( Mode = dmOnlyDoc1 ) then M := FlagModifiedFile1;
  FFlags := FFlags and (not M);
  for I:=1 to FUndoRedoBuffer.UndoCount do begin
    TDiffUndoItem( FUndoRedoBuffer.UndoItems[I-1] ).Flags := M or TDiffUndoItem( FUndoRedoBuffer.UndoItems[I-1] ).Flags;
  end;
  for I:=1 to FUndoRedoBuffer.RedoCount do begin
    TDiffUndoItem( FUndoRedoBuffer.RedoItems[I-1] ).Flags := M or TDiffUndoItem( FUndoRedoBuffer.RedoItems[I-1] ).Flags;
  end;

  ModificationFlagsChanged;

  Result := true;
end;

procedure TMainForm.CmdSaveLeftFileClick(Sender: TObject);
begin
  if( FFileName1 = '' ) then
    CmdSaveLeftFileAsClick(Sender)
  else
    SaveFile( FFileName1, dmOnlyDoc1 );
end;

procedure TMainForm.CmdSaveRightFileClick(Sender: TObject);
begin
  if( FFileName2 = '' ) then
    CmdSaveRightFileAsClick(Sender)
  else
    SaveFile( FFileName2, dmOnlyDoc2 );
end;

procedure TMainForm.BtnSaveAllClick(Sender: TObject);
begin
  CmdSaveLeftFileClick(Sender);
  CmdSaveRightFileClick(Sender);
end;

procedure TMainForm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  CanClose := AllowCloseCurrentDocument;
end;

procedure TMainForm.CmdSaveLeftFileAsClick(Sender: TObject);
begin
  if( SaveDialogLeft.InitialDir = '' ) then begin
    SaveDialogLeft.InitialDir := ExtractFilePath( FFilename1 );
  end;
  SaveDialogLeft.FileName := FFilename1;
  if( SaveDialogLeft.Execute ) then begin
    if SaveFile( SaveDialogLeft.FileName, dmOnlyDoc1 ) then begin
      FFilename1 := SaveDialogLeft.FileName;
    end;
  end;
end;

procedure TMainForm.CmdSaveRightFileAsClick(Sender: TObject);
begin
  if( SaveDialogRight.InitialDir = '' ) then begin
    SaveDialogRight.InitialDir := ExtractFilePath( FFilename2 );
  end;
  SaveDialogRight.FileName := FFilename2;
  if( SaveDialogRight.Execute ) then begin
    if SaveFile( SaveDialogRight.FileName, dmOnlyDoc2 ) then begin
      FFilename2 := SaveDialogRight.FileName;
      UpdateFormCaption;
    end;
  end;
end;

procedure TMainForm.HandleDoubleClick( Sender: TPaintBox; ModeFrom, ModeTo: TDiffLineMode );
var
  P: TPoint;
  L: integer;
begin
  // Note: in theory we need the mouse position when it double clicked,
  // however this is enough for this program!
  GetCursorPos( P );
  P := Sender.ScreenToClient( P );

  // Get the index of the line under the cursor
  L := (P.Y - 2) div GetLineHeight + VertScrollBar.Position;
  if( L >= 0 )and( L < FDiffDocument.Count ) then begin
    if( FDiffDocument.Mode[L] = ModeFrom ) then begin
      ChangeLineBlock( L, ModeFrom, ModeTo, true );
    end;
  end;
end;

procedure TMainForm.PbxView1DblClick(Sender: TObject);
begin
  HandleDoubleClick( PbxView1, dmOnlyDoc1, dmCopiedFromDoc1 );
end;

procedure TMainForm.PbxView2DblClick(Sender: TObject);
var
  ModeTo: TDiffLineMode;
begin
  if( FOptions.DblClickOnRightFileDeletes.Value ) then
    ModeTo := dmDeletedFromDoc2
  else
    ModeTo := dmCopiedFromDoc2;
  HandleDoubleClick( PbxView2, dmOnlyDoc2, ModeTo );
end;

procedure TMainForm.CmdPreferencesClick(Sender: TObject);
begin
  OptionsForm := TOptionsForm.Create(Self);
  OptionsForm.SetOptions( FOptions );
  if( OptionsForm.ShowModal = mrOk ) then begin
    OptionsForm.GetOptions( FOptions );
    FormPaint( Self );
  end;
  OptionsForm.Free;
end;

// Reload a recent file difference operation
procedure TMainForm.CmdRecentFilesClick(Sender: TObject);
var
  M: TMenuItem;
  P: integer;
  S, N1, N2: string;
procedure RemoveBadItem;
  begin
    FOptions.HistoryRecent.Strings.Delete( M.Tag );
    UpdateRecentFilesMenu;
  end;
begin
  M := Sender as TMenuItem;
  S := FOptions.HistoryRecent.Strings[ M.Tag ];
  P := Pos( '|', S );
  if( P = 0 ) then begin
    RemoveBadItem;
    Exit;
  end;
  Delete( S, 1, P );
  P := Pos( '|', S );
  if( P = 0 ) then begin
    RemoveBadItem;
    Exit;
  end;
  N1 := Copy( S, 1, P-1 );
  N2 := Copy( S, P+1, Length(S) );
  if( N1 = '' )or( N2 = '' ) then begin
    RemoveBadItem;
    Exit;
  end;
  LoadFiles( N1, N2 );
end;

procedure TMainForm.CmdRefreshClick(Sender: TObject);
var
  N: integer;
begin
  // Refresh
  N := VertScrollBar.Position;
  LoadFiles( '', '' );
  SetViewPosition( N );
end;

procedure TMainForm.CmdAutosynchClick(Sender: TObject);
var
  I, J: integer;
  OldMode, NewMode: TDiffLineMode;
  U: TUndoRedoItemList;
  M: integer;
begin
  if( AutosynchForm.ShowModal <> mrOk ) then
    Exit;

  // Set synch mode
  M := 0;
  if( AutosynchForm.RadioButton2.Checked ) then
    M := 1
  else if( AutosynchForm.RadioButton3.Checked ) then
    M := 2
  else if( AutosynchForm.RadioButton4.Checked ) then
    M := 3;

  // Since this operation can change a number of blocks,
  // we use a list to hold all the changes
  U := TUndoRedoItemList.Create;

  // Scan the active lines list to look for differences
  I := 0;
  while( I < FDiffDocument.Count ) do begin
    if IsDifference( FDiffDocument.Mode[I] ) then begin
      J := GetBlockEnd( I ) + 1;

      OldMode := FDiffDocument.Mode[I];
      NewMode := OldMode;
      if( OldMode = dmOnlyDoc1 ) then begin
        if( M = 0 )or( M = 1 ) then
          NewMode := dmCopiedFromDoc1
        else if( M = 2 ) then
          NewMode := dmDeletedFromDoc1;
      end
      else begin
        if( M = 2 )or( M = 3 ) then
          NewMode := dmCopiedFromDoc2
        else if( M = 0 ) then
          NewMode := dmDeletedFromDoc2;
      end;

      if( NewMode <> OldMode ) then begin
        U.Add( TDiffUndoItem.Create( OldMode, NewMode, I, J, -1 ) );

        // Change the block
        while( I < J ) do begin
          FDiffDocument.Line[I].Mode := NewMode;
          Inc( I );
        end;
      end;

      I := J-1;
    end;

    Inc( I );
  end;

  // Did we change the document?
  if( U.Count > 0 ) then begin
    // Tag the last item in the list with the proper flags and push it
    TDiffUndoItem( U.Item[ U.Count-1 ] ).Flags := FFlags;
    FUndoRedoBuffer.Push( U );

    // This operation only changes the second file
    FFlags := FFlags or FlagModifiedFile2;

    // Update user interface
    ModificationFlagsChanged;
    DiffLineListChanged;
  end
  else begin
    // Did nothing, free resources and exit
    U.Free;
  end;
end;

procedure TMainForm.ShowFileProperties( SelectRightFile: boolean );
var
  FP: TPropertiesForm;
begin
  FP := TPropertiesForm.Create( Self );
  FP.SetFiles( FFilename1, FFilename2, SelectRightFile );
  FP.ShowModal;
  FP.Free;
end;

procedure TMainForm.PopPropertiesLeftClick(Sender: TObject);
begin
  ShowFileProperties(false);
end;

procedure TMainForm.PopPropertiesRightClick(Sender: TObject);
begin
  ShowFileProperties(true);
end;

procedure TMainForm.DisplayHint(Sender: TObject);
begin
  StatusBar.SimpleText := GetLongHint(Application.Hint);
end;

procedure TMainForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  case Key of
  VK_HOME: SetViewPosition( 0 );
  VK_END: SetViewPosition( VertScrollBar.Max );
  VK_DOWN: SetViewPosition( VertScrollBar.Position + 1 );
  VK_UP: SetViewPosition( VertScrollBar.Position - 1 );
  VK_NEXT: SetViewPosition( VertScrollBar.Position + VertScrollBar.LargeChange );
  VK_PRIOR: SetViewPosition( VertScrollBar.Position - VertScrollBar.LargeChange );
  VK_LEFT: if( HorzScrollBar.Position > HorzScrollBar.Min ) then HorzScrollBar.Position := HorzScrollBar.Position - 1;
  VK_RIGHT: if( HorzScrollBar.Position < HorzScrollBar.Max ) then HorzScrollBar.Position := HorzScrollBar.Position + 1;
  VK_F1: OpenHelp( 'main' );
  VK_F6: begin Width := 800; Height := 600; Caption := IntToStr(Width); end;  
  end;
end;

procedure TMainForm.BtnShowLineNumbersClick(Sender: TObject);
begin
  FOptions.ShowLineNumbers.Value := not FOptions.ShowLineNumbers.Value;
  FormPaint(Self);
end;

(*--------------------------------------------------
    Drag and drop support
--------------------------------------------------*)
type
  TBlockDragObject = class(TDragObject)
  public
    LoLine, HiLine, Source: integer;
    constructor Create( ALoLine, AHiLine, ASource: integer );
  end;

constructor TBlockDragObject.Create;
begin
  LoLine := ALoLine;
  HiLine := AHiLine;
  Source := ASource;
end;

const
  FDragLineView: integer = -1;

procedure TMainForm.PbxView1StartDrag(Sender: TObject;
  var DragObject: TDragObject);
var
  L, H: integer;
begin
  L := FDragLineView;
  if( L >= 0 ) then begin
    H := L;
    while( L > 0 )and( FDiffDocument.Mode[L-1] = dmOnlyDoc1 ) do
      Dec( L );
    while( H < FDiffDocument.Count )and( FDiffDocument.Mode[H] = dmOnlyDoc1 ) do
      Inc( H );
    DragObject := TBlockDragObject.Create( L, H, 1 );
  end;
end;

procedure TMainForm.PbxView2StartDrag(Sender: TObject;
  var DragObject: TDragObject);
var
  L, H: integer;
begin
  L := FDragLineView;
  if( L >= 0 ) then begin
    H := L;
    while( L > 0 )and( FDiffDocument.Mode[L-1] = dmOnlyDoc2 ) do
      Dec( L );
    while( H < FDiffDocument.Count )and( FDiffDocument.Mode[H] = dmOnlyDoc2 ) do
      Inc( H );
    DragObject := TBlockDragObject.Create( L, H, 2 );
  end;
end;

procedure TMainForm.PbxView1MouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  L: integer;
begin
  if( Button <> mbLeft )or( Shift <> [ssLeft] ) then
    Exit;
  L := (Y - 2) div GetLineHeight + VertScrollBar.Position;
  if( L < 0 )or( L >= FDiffDocument.Count ) then
    Exit;
  if( FDiffDocument.Mode[L] = dmOnlyDoc1 ) then begin
    FDragLineView := L;
    PbxView1.BeginDrag( false );
  end;
end;

procedure TMainForm.PbxView2MouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  L: integer;
begin
  if( Button <> mbLeft )or( Shift <> [ssLeft] ) then
    Exit;
  L := (Y - 2) div GetLineHeight + VertScrollBar.Position;
  if( L < 0 )or( L >= FDiffDocument.Count ) then
    Exit;
  if( FDiffDocument.Mode[L] = dmOnlyDoc2 ) then begin
    FDragLineView := L;
    PbxView2.BeginDrag( false );
  end;
end;

function TMainForm.ValidateDragOver( Source: TObject; Y: integer ): boolean;
var
  L: integer;
  O: TBlockDragObject;
begin
  Result := false;
  if( not IsDragObject(Source) ) then
    Exit;
  L := (Y - 2) div GetLineHeight + VertScrollBar.Position;
  if( L < 0 )or( L >= FDiffDocument.Count ) then
    Exit;
  O := Source as TBlockDragObject;
  if( L < O.LoLine )or( L >= O.HiLine ) then
    Exit;
  Result := true;
end;

procedure TMainForm.PbxView1DragOver(Sender, Source: TObject; X,
  Y: Integer; State: TDragState; var Accept: Boolean);
begin
  Accept := ValidateDragOver( Source, Y );
end;

procedure TMainForm.PbxView2DragOver(Sender, Source: TObject; X,
  Y: Integer; State: TDragState; var Accept: Boolean);
begin
  Accept := ValidateDragOver( Source, Y );
end;

procedure TMainForm.ExecuteDragDrop( Source: TObject; SourceView: integer; ModeFrom, ModeTo: TDiffLineMode );
var
  O: TBlockDragObject;
begin
  if( not IsDragObject(Source) ) then
    Exit;
  O := Source as TBlockDragObject;
  if( O.Source <> SourceView ) then
    Exit;
  ChangeLineBlock( O.LoLine, ModeFrom, ModeTo, true );
end;

procedure TMainForm.PbxView1DragDrop(Sender, Source: TObject; X,
  Y: Integer);
begin
  ExecuteDragDrop( Source, 2, dmOnlyDoc2, dmCopiedFromDoc2 );
end;

procedure TMainForm.PbxView2DragDrop(Sender, Source: TObject; X,
  Y: Integer);
begin
  ExecuteDragDrop( Source, 1, dmOnlyDoc1, dmCopiedFromDoc1 );
end;

(*--------------------------------------------------
    Differences browsing
--------------------------------------------------*)
function TMainForm.IndexOfPrevDifference: integer;
begin
  Result := VertScrollBar.Position - 1;
  while( Result >= 0 )and( not IsDifference(FDiffDocument.Mode[Result]) ) do
    Dec( Result );
  if( Result >= 0 ) then
    Result := GetBlockStart(Result);
end;

function TMainForm.IndexOfNextDifference: integer;
begin
  Result := GetBlockEnd( VertScrollBar.Position ) + 1;
  if( Result < FDiffDocument.Count )and( not IsDifference(FDiffDocument.Mode[Result]) ) then
    Result := GetBlockEnd( Result ) + 1;
  if( Result >= FDiffDocument.Count ) then
    Result := -1;
end;

procedure TMainForm.UpdateDifferenceNavigators;
begin
  BtnPrevDiff.Enabled := IndexOfPrevDifference >= 0;
  CmdPreviousDifference.Enabled := BtnPrevDiff.Enabled;
  BtnNextDiff.Enabled := IndexOfNextDifference >= 0;
  CmdNextDifference.Enabled := BtnNextDiff.Enabled;
  CmdAutosynch.Enabled := (FDiffDocument.DiffCount1+FDiffDocument.DiffCount2) <> 0;
  BtnMerge.Enabled := CmdAutosynch.Enabled;
end;

procedure TMainForm.CmdPreviousDifferenceClick(Sender: TObject);
var
  L: integer;
begin
  L := IndexOfPrevDifference;
  if( L >= 0 ) then
      SetViewPosition( L );
end;

procedure TMainForm.CmdNextDifferenceClick(Sender: TObject);
var
  L: integer;
begin
  L := IndexOfNextDifference;
  if( L >= 0 ) then
      SetViewPosition( L );
end;

(*--------------------------------------------------
    Full text search
--------------------------------------------------*)
procedure TMainForm.FindNext;
var
  MatchCase: boolean;
  SearchText, SearchLine: string;
  Line: TDiffLine;
  N, M: integer;
begin
  SearchText := FindDialog.FindText;
  if( SearchText = '' ) then
    Exit;
  MatchCase := frMatchCase in FindDialog.Options;
  if( not MatchCase ) then
    SearchText := UpperCase(SearchText);
  repeat
    if( frDown in FindDialog.Options ) then
      Inc( FCurrentLine )
    else
      Dec( FCurrentLine );
    Line := FDiffDocument.Line[ FCurrentLine ];
    if( Line <> nil ) then begin
      if( Line.Mode = dmOnlyDoc1 ) then
        SearchLine := Line.Line1
      else
        SearchLine := Line.Line2;
      if( not MatchCase ) then
        SearchLine := Uppercase(SearchLine);
      if( Pos( SearchText, SearchLine ) > 0 ) then begin
        // Found
        N := FCurrentLine - GetVisibleLines div 2;
        M := VertScrollBar.Position;
        SetViewPosition( N );
        if( VertScrollBar.Position = M ) then begin
          FormPaint(Self);
        end;
        Exit;
      end;
    end;
  until( Line = nil );
  FCurrentLine := -1;
  StatusBar.SimpleText := ' Search text ''' + FindDialog.FindText + ''' not found';
  FormPaint(Self);
end;

procedure TMainForm.CmdFindAgainClick(Sender: TObject);
begin
  FindNext;
end;

procedure TMainForm.FindDialogFind(Sender: TObject);
begin
  FindDialog.CloseDialog;
  FCurrentLine := VertScrollBar.Position;
  if( frDown in FindDialog.Options ) then
    Dec( FCurrentLine )
  else
    Inc( FCurrentLine );
  FindNext;
end;

procedure TMainForm.CmdFindClick(Sender: TObject);
begin
  FindDialog.Execute;
end;

{$WARNINGS OFF}
procedure TMainForm.PbxView2MouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
var
  L: integer;
  F: boolean;
  P: TPoint;
  B: TPaintBox;
  A: TAutoPopupMode;
begin
  // Exit now if popping up is disabled
  if( FOptions.PopupMargin.Value <= 0 ) then
    Exit;

  B := Sender as TPaintBox;

  F := false;
  if( X >= B.Width - FOptions.PopupMargin.Value ) then begin
    L := (Y - 2) div GetLineHeight + VertScrollBar.Position;
    if( B = PbxView1 ) then begin
      F := FDiffDocument.Mode[L] in [dmOnlyDoc1];
      A := pmDoc1;
    end
    else begin
      F := FDiffDocument.Mode[L] in [dmOnlyDoc2];
      A := pmDoc2;
    end;
  end;

  if( F ) then begin
    if( FormDiffAutoPopup.Visible ) then
      Exit;
    P := Point( B.Left + B.Width - FormDiffAutoPopup.Width - 3, B.Top + Y - 16 );
    FormDiffAutoPopup.Popup( ClientToScreen(P), A, L )
  end
  else begin
    FormDiffAutoPopup.Hide;
  end;
end;
{$WARNINGS ON}

procedure TMainForm.CmdOpenHelpClick(Sender: TObject);
begin
  OpenHelp( 'index' );
end;

procedure TMainForm.CmdShortcutsClick(Sender: TObject);
begin
  OpenHelp( 'shortcuts' );
end;

procedure TMainForm.BtnShowSeparatorsClick(Sender: TObject);
begin
  FOptions.ShowBlockSeparators.Value := not FOptions.ShowBlockSeparators.Value;
  FormPaint(Self);
end;

procedure TMainForm.BtnSyntaxHilightClick(Sender: TObject);
begin
  FOptions.SyntaxHighlight.Value := not FOptions.SyntaxHighlight.Value;
  FormPaint(Self);
end;

procedure TMainForm.FormKeyPress(Sender: TObject; var Key: Char);
var
  L: integer;
  P: TPoint;
  OldMode: TDiffLineMode;
  NewMode: TDiffLineMode;
  WholeBlock: boolean;
function HandleSimpleCommand: boolean;
  begin
    Result := true;
    case Key of
      'e','E': PopEditLeftClick(Self);
      'r','R': PopEditRightClick(Self);
      '<': CmdPreviousDifferenceClick(Self);
      '>': CmdNextDifferenceClick(Self);
      'a','A': SetViewPosition( VertScrollBar.Position + VertScrollBar.LargeChange );
      else Result := false;
    end;
  end;
begin
  if( HandleSimpleCommand() ) then begin
    Exit;
  end;

  // Find whether the mouse is inside the document window
  GetCursorPos( P );
  P := ScreenToClient( P );
  if PtInRect( PbxView1.BoundsRect, P ) or PtInRect( PbxView2.BoundsRect, P ) then begin
    L := (P.Y - PbxView1.Top - 2) div GetLineHeight + VertScrollBar.Position;

    if( L < 0 )or( L >= FDiffDocument.Count ) then
      Exit;

    OldMode := FDiffDocument.Mode[L];
    NewMode := dmNone;
    WholeBlock := true;

    // Exit now if the cursor is over a "grayed" area
    if( OldMode = dmOnlyDoc1 )and not PtInRect( PbxView1.BoundsRect, P ) then
      Exit;
    if( OldMode = dmOnlyDoc2 )and not PtInRect( PbxView2.BoundsRect, P ) then
      Exit;

    case Key of
      '1': case OldMode of // Merge 1 over 2
        dmOnlyDoc1: NewMode := dmCopiedFromDoc1;
        dmOnlyDoc2: NewMode := dmDeletedFromDoc2;
      end;
      '2': case OldMode of // Merge 2 over 1
        dmOnlyDoc1: NewMode := dmDeletedFromDoc1;
        dmOnlyDoc2: NewMode := dmCopiedFromDoc2;
      end;
      'c','C': case OldMode of // Copy
        dmOnlyDoc1: NewMode := dmCopiedFromDoc1;
        dmOnlyDoc2: NewMode := dmCopiedFromDoc2;
      end;
      'd','D': case OldMode of // Delete
        dmOnlyDoc1: NewMode := dmDeletedFromDoc1;
        dmOnlyDoc2: NewMode := dmDeletedFromDoc2;
      end;
      'z','Z': case OldMode of // Block undo
        dmCopiedFromDoc1: NewMode := dmOnlyDoc1;
        dmCopiedFromDoc2: NewMode := dmOnlyDoc2;
      end;
    end;

    if( NewMode <> dmNone ) then begin
      ChangeLineBlock( L, OldMode, NewMode, WholeBlock );
    end;
  end;
end;

function TMainForm.GetActualDocument( Mode: TDiffLineMode ): TStringList;
var
  I: integer;
  L: TDiffLine;
begin
  Result := TStringList.Create;
  for I:=1 to FDiffDocument.Count do begin
    L := FDiffDocument.Line[I-1];
    if( Mode = dmOnlyDoc1 )and( L.Mode in [dmOnlyDoc1,dmBothDocuments,dmCopiedFromDoc1,dmCopiedFromDoc2] ) then
      Result.Add( L.Line1 )
    else if( Mode = dmOnlyDoc2 )and( L.Mode in [dmOnlyDoc2,dmBothDocuments,dmCopiedFromDoc1,dmCopiedFromDoc2] ) then
      Result.Add( L.Line2 );
  end;
end;

function TMainForm.EditDoc( M: TDiffLineMode ): boolean;
var
  L: TStringList;
  D: TList;
  IgnoreBlanks: boolean;
  IgnoreCase: boolean;
begin
  Result := false;

  if( M = dmOnlyDoc1 ) then begin
    EditorForm.SyntaxTokenizer := FSyntaxTokenizer[0];
    EditorForm.SetCaption( FFilename1 );
    L := GetActualDocument( dmOnlyDoc1 );
  end
  else begin
    EditorForm.SyntaxTokenizer := FSyntaxTokenizer[1];
    EditorForm.SetCaption( FFilename2 );
    L := GetActualDocument( dmOnlyDoc2 );
  end;

  EditorForm.TabSize := FOptions.TabSize.Value;
  EditorForm.Lines := L; // Note: this is a property, string list is actually copied!

  if( EditorForm.ShowModal = mrOk ) then begin
    D := TList.Create;

    // Get the difference between the original and the edited document
    // (note we temporarily override the diff settings to get a "true"
    // diff between the original and the edited document)
    IgnoreBlanks := FDiffDocument.DiffEngine.IgnoreBlanks;
    IgnoreCase := FDiffDocument.DiffEngine.IgnoreCase;
    FDiffDocument.DiffEngine.IgnoreBlanks := false;
    FDiffDocument.DiffEngine.IgnoreCase := false;

    FDiffDocument.DiffEngine.Diff( L, EditorForm.Lines, D );

    FDiffDocument.DiffEngine.IgnoreBlanks := IgnoreBlanks;
    FDiffDocument.DiffEngine.IgnoreCase := IgnoreCase;

    TDiffEngine.MakeSimpleDiffList( L, EditorForm.Lines, D );

    if( D.Count > 0 ) then begin
      // Document has been modified, save diff list in the undo buffer
      FUndoRedoBuffer.Push( TEditDiffDocUndoItem.Create( D, M, FFlags ) );

      // Change the document
      if( M = dmOnlyDoc1 ) then begin
        FDiffDocument.RawDoc1.Assign( EditorForm.Lines );
        FFlags := FFlags or FlagModifiedFile1;
      end
      else begin
        FDiffDocument.RawDoc2.Assign( EditorForm.Lines );
        FFlags := FFlags or FlagModifiedFile2;
      end;
      
      FDiffDocument.RebuildAll;

      // Update user interface
      ModificationFlagsChanged;
      DiffLineListChanged;

      Result := true;
    end
    else begin
      // No modifications, free resources and exit
      D.Free;
    end;
  end;

  L.Free;
end;

procedure TMainForm.PopEditLeftClick(Sender: TObject);
begin
  EditDoc( dmOnlyDoc1 );
end;

procedure TMainForm.PopEditRightClick(Sender: TObject);
begin
  EditDoc( dmOnlyDoc2 );
end;

end.

