(*
    LCS - A study for a simple differencing program
    Persistent options
    Copyright (c) 2002-2003,2004 Alessandro Scotti
    www.ascotti.org

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)
unit Options;

interface

uses
  Config, Graphics;

const
  DefaultColorLineNumber = clInfoBk;
  DefaultColorOnlyDoc1 = $00B0E0B0;
  DefaultColorOnlyDoc2 = $00B0B0FF;
  DefaultColorDefault = clWindow;
  DefaultColorMissingLine = clGray;
  DefaultColorChangedLine = $00F0DCBE;

  DefaultTabSize = 4;
  DefaultPopupMargin = 16;

type
  TAppOptions = record
    ColorLineNumber: TConfigInteger;
    ColorOnlyDoc1: TConfigInteger;
    ColorOnlyDoc2: TConfigInteger;
    ColorDefault: TConfigInteger;
    ColorMissingLine: TConfigInteger;
    ColorChangedLine: TConfigInteger;
    IgnoreBlanks: TConfigBoolean;
    IgnoreCase: TConfigBoolean;
    BackupMode: TConfigInteger;
    DblClickOnRightFileDeletes: TConfigBoolean;
    RestoreWindowSettings: TConfigBoolean;
    WindowPosX: TConfigInteger;
    WindowPosY: TConfigInteger;
    WindowWidth: TConfigInteger;
    WindowHeight: TConfigInteger;
    WindowIsMaximized: TConfigBoolean;
    ShowBlockSeparators: TConfigBoolean;
    ShowLineNumbers: TConfigBoolean;
    SyntaxHighlight: TConfigBoolean;
    TabSize: TConfigInteger;
    PopupMargin: TConfigInteger;
    OpenFileFilters: TConfigString;
    HistoryFile1: TConfigStringList; // History of loaded "left" files
    HistoryFile2: TConfigStringList; // History of loaded "right" files
    HistoryRecent: TConfigStringList; // History of recent comparisons
  end;

  TLanguageOptions = class
    FName: string;
    FIgnoreCase: TConfigBoolean;
    FQuoteChars: TConfigString;
    FQuoteEscape: TConfigString;
    FLineComment: TConfigString;
    FCommentBegin: TConfigString;
    FCommentEnd: TConfigString;
    FIdentifiers: TConfigString;
    FExtensions: TConfigString;
    FConfigGroup: TConfigGroup;
  public
    constructor Create( AName: string );
    destructor Destroy; override;

    function IsValid: boolean;

    property Name: string read FName;
    property IgnoreCase: TConfigBoolean read FIgnoreCase;
    property QuoteChars: TConfigString read FQuoteChars;
    property QuoteEscape: TConfigString read FQuoteEscape;
    property LineComment: TConfigString read FLineComment;
    property CommentBegin: TConfigString read FCommentBegin;
    property CommentEnd: TConfigString read FCommentEnd;
    property Identifiers: TConfigString read FIdentifiers;
    property Extensions: TConfigString read FExtensions;
    property ConfigGroup: TConfigGroup read FConfigGroup;
  end;

implementation

function TLanguageOptions.IsValid: boolean;
begin
  Result := (FIdentifiers.Value <> '')and(FExtensions.Value <> '');
end;

constructor TLanguageOptions.Create( AName: string );
begin
  FName := AName;
  FIgnoreCase := TConfigBoolean.Create( 'IgnoreCase', false );
  FQuoteChars := TConfigString.Create( 'QuoteChars', '' );
  FQuoteEscape := TConfigString.Create( 'QuoteEscapeChar', '' );
  FLineComment := TConfigString.Create( 'EndOfLineComment', '' );
  FCommentBegin := TConfigString.Create( 'CommentBegin', '' );
  FCommentEnd := TConfigString.Create( 'CommentEnd', '' );
  FIdentifiers := TConfigString.Create( 'Identifiers', '' );
  FExtensions := TConfigString.Create( 'Extensions', '' );
  FConfigGroup := TConfigGroup.Create( AName );
  FConfigGroup.Add( FIgnoreCase );
  FConfigGroup.Add( FQuoteChars );
  FConfigGroup.Add( FQuoteEscape );
  FConfigGroup.Add( FLineComment );
  FConfigGroup.Add( FCommentBegin );
  FConfigGroup.Add( FCommentEnd );
  FConfigGroup.Add( FIdentifiers );
  FConfigGroup.Add( FExtensions );
end;

destructor TLanguageOptions.Destroy;
begin
  FConfigGroup.Free;
end;

end.
