(*
    Misc functions
    Copyright (c) 2002-2003,2004 Alessandro Scotti
    www.ascotti.org

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)
unit Utils;

interface

function BoolToInteger( B: boolean ): integer;
function ChooseInteger( B: boolean; VTrue, VFalse: integer ): integer;
function ExpandTabs( const L: string; TabSize: integer ): string;
function StrRChr( C: char; const S: string ): integer;
function ShortenFilePath( FullName: string; MaxLen: integer ): string;
function ExtractToken( var S: string; Delimiter: char ): string;

procedure OpenHelp( Page: string );

implementation

uses
  SysUtils, ShellAPI, Forms, Windows;

function BoolToInteger( B: boolean ): integer;
begin
  Result := ChooseInteger( B, 1, 0 );
end;

function ChooseInteger( B: boolean; VTrue, VFalse: integer ): integer;
begin
  if( B ) then Result := VTrue else Result := VFalse;
end;

function StrRChr( C: char; const S: string ): integer;
var
  I: integer;
begin
  Result := 0;
  for I:=Length(S) downto 0 do begin
    if( S[I] = C ) then begin
      Result := I;
      break;
    end;
  end;
end;

// Shortens a file path so that it takes less space on display
function ShortenFilePath( FullName: string; MaxLen: integer ): string;
var
  N, P, R: string;
  I: integer;
begin
  Result := FullName;
  if( Length(FullName) > MaxLen ) then begin
    N := ExtractFileName( FullName );
    P := ExtractFilePath( FullName );
    I := Pos( '\', P );
    if( I > 0 ) then begin
      R := Copy( P, 1, I );
      Delete( P, 1, I );
      while( true ) do begin
        I := Pos( '\', P );
        if( I = 0 ) then
          break;
        if( (Length(R) + Length(N) + I) > MaxLen ) then
          break;
        R := R + Copy( P, 1, I );
        Delete( P, 1, I );
      end;
      if( P <> '' ) then
        R := R + '...\';
      Result := R + N;
    end
  end;
end;

function ExpandTabs( const L: string; TabSize: integer ): string;
var
  P: integer;
begin
  Result := L;
  P := Pos( #9, L );
  if( P > 0 ) then begin
    repeat
      Delete( Result, P, 1 );
      Insert( StringOfChar(' ',TabSize-((P-1) mod Tabsize)), Result, P );
      repeat
        Inc( P );
      until( Result[P] = #9 )or( Result[P] = #0 );
    until( Result[P] = #0 );
  end;
end;

procedure OpenHelp( Page: string );
begin
  Page := ExtractFilePath(Application.ExeName) + 'help\' + Page + '.htm';
  ShellExecute( Application.MainForm.Handle, 'open', PChar(Page), nil, nil, SW_NORMAL );
end;

function ExtractToken( var S: string; Delimiter: char ): string;
var
  P: integer;
begin
  P := Pos( Delimiter, S );
  if( P > 0 ) then begin
    Result := Copy( S, 1, P-1 );
    Delete( S, 1, P );
  end
  else begin
    Result := S;
    S := '';
  end;
end;

end.
